"""Simple functions relating to cartesian grid boxes

A box is a logical cuboid subset of the cells of a cartesian grid.
A box is defined by a small numpy array: [[min_k, min_j, min_i], [max_k, max_j, max_i]].
The cells identified by the max indices are included in the box (not following the python convention)
The ordering of the i,j & k indices might be reversed - identifier names then have a suffix of _ijk instead of _kji.
The indices can be in simulator convention, starting at 1, or python convention, starting at 0, indicated by suffix of 0 or 1
"""

version = '13th July 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np


def extent_of_box(box):  # returns 3 element extent of box (box can be kji or ijk, 0 or 1 based)
    """Returns a 3 integer numpy array holding the size of the box, with the same ordering as the box.

    input argument (unmodified):
       box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid

    returns: numpy int array of shape (3)
          the extent (shape) of the cuboid defined by box
    """

    assert box.ndim == 2 and box.shape == (2, 3)
    return box[1] - box[0] + 1  # numpy array operation


def volume_of_box(box):
    """Returns the number of cells in the logical 3D cell space defined by box.

    input argument (unmodified):
       box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid

    returns: int
          the total number of cells in box
    """

    return (box[1, 0] - box[0, 0] + 1) * (box[1, 1] - box[0, 1] + 1) * (box[1, 2] - box[0, 2] + 1)


def central_cell(box):
    """Returns the indices of the cell at the centre of the box."""
    return box[0] + ((box[1] - box[0]) // 2)


def string_iijjkk1_for_box_kji0(box_kji0):
    """Returns a string representing the box space in simulator protocol, eg. '[1:5, 3:20, 100:103]'.

    input argument (unmodified):
       box_kji0: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          with python kji ordering and zero start for indices

    returns: string
       human readable representation of box in Fortran/simulator ijk protocol starting 1
    """

    return '[' + str(box_kji0[0, 2] + 1) + ':' + str(box_kji0[1, 2] + 1) + ', ' +  \
                 str(box_kji0[0, 1] + 1) + ':' + str(box_kji0[1, 1] + 1) + ', ' +  \
                 str(box_kji0[0, 0] + 1) + ':' + str(box_kji0[1, 0] + 1) + ']'


def spaced_string_iijjkk1_for_box_kji0(box_kji0, colon_separator = ' '):
    """Returns a string representing the box space in simulator input format, eg. '1 5  3 20  100 103'.

    input arguments (unmodified):
       box_kji0: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          with python kji ordering and zero start for indices
       colon_separator: string (typically ':' or ' ')
          the character(s) included in the return string between lower and upper bounds in each direction

    returns: string
       ascii representation of box in Fortran/simulator ijk protocol starting 1, suitable for use in include files
    """

    return str(box_kji0[0, 2] + 1) + colon_separator + str(box_kji0[1, 2] + 1) + '  ' +  \
           str(box_kji0[0, 1] + 1) + colon_separator + str(box_kji0[1, 1] + 1) + '  ' +  \
           str(box_kji0[0, 0] + 1) + colon_separator + str(box_kji0[1, 0] + 1)


def box_kji0_from_words_iijjkk1(words):
    """Returns an integer array of extent [2, 3] converted from a list of words representing logical box.

    input argument (unmodified):
       words: a list of strings with at least 6 elements castable to int
          [min_i, max_i, min_j, max_j, min_k, max_k] in Fortran/simulator protocol (indices start at 1)

    returns: 2D numpy int array of shape (2, 3)
       [min, max][k, j, i] with cell indices in python protocol (zero base)

    notes:
       designed to take string format numbers: minI maxI minJ maxJ minK maxK
       and convert to a pair of integer cell id triplets: min(k, j, i), max(k, j, i)
       NB: output indices have been decremented by 1 (for python indexing starting at zero)
    """

    assert len(words) >= 6  # expecting minI maxI minJ maxJ minK maxK value
    box = np.zeros([2, 3], dtype = 'int')
    box[0, 0] = int(words[4]) - 1  # k min
    box[1, 0] = int(words[5]) - 1  # k max
    box[0, 1] = int(words[2]) - 1  # j min
    box[1, 1] = int(words[3]) - 1  # j max
    box[0, 2] = int(words[0]) - 1  # i min
    box[1, 2] = int(words[1]) - 1  # i max
    assert box[0, 0] <= box[1, 0] and box[0, 1] <= box[1, 1] and box[0, 2] <= box[1, 2]

    return box


def cell_in_box(cell, box):
    """Returns True if cell is within box, otherwise False.

    input arguments (unmodified):
       cell: numpy int array of shape (3)
          index of a cell in a 3D cartesian grid, in the same protocol as box (usually python protocol kji, zero base)
       box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          in the same protocol as cell

    returns: boolean
       True if cell is within box, False otherwise
    """

    return (box[0, 0] <= cell[0] <= box[1, 0]) and (box[0, 1] <= cell[1] <= box[1, 1]) and (box[0, 2] <= cell[2] <=
                                                                                            box[1, 2])


def valid_box(box, host_extent):
    """Returns True if the entire box is within a grid of size host_extent.

    input arguments (unmodified):
       box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          in python protocol of zero base, kji (normally) or ijk ordering same as for host_extent
       host_extent: triple int
          the extent (shape) of a 3D cartesian grid

    returns: boolean
       True if box is a valid box within a grid of shape host_extent, False otherwise
    """

    if box.ndim != 2 or box.shape != (2, 3) or box.dtype != 'int':
        return False
    if len(host_extent) != 3:
        return False
    for d in range(3):
        if box[0, d] < 0 or box[0, d] > box[1, d] or box[1, d] >= host_extent[d]:
            return False
    return True


def single_cell_box(cell):
    """Returns a box containing the single given cell; protocol for box matches that of cell.

    input argument (unmodified):
       cell: numpy int array of shape (3)
          indices of a cell within a 3D cartesian grid, usually in python protocol (kji ordering, zero base)

    returns: numpy int array of shape (2, 3)
       indices defining a minimal box containing a single cell; protocol is same as that of cell
    """

    assert cell.ndim == 1 and cell.size == 3
    box = np.zeros((2, 3), dtype = 'int')
    box[0] = cell  # numpy 1D array op
    box[1] = cell  # numpy 1D array op
    return box


def full_extent_box0(extent):
    """Returns a box containing all the cells in a grid of the given extent.

    input argument (unmodified):
       extent: numpy int array of shape (3)
          extent (shape) of a 3D cartesian grid, usually in kji python protocol

    returns: numpy int array of shape (2, 3)
       indices defining a maximal box containing the entire grid; kji ordering is same as that of extent; zero base
    """

    assert extent.ndim == 1 and extent.size == 3
    box = np.zeros((2, 3), dtype = 'int')
    box[1, :] = extent - 1  # numpy 1D array op
    return box


def union(box_1, box_2):
    """Returns the box which contains both box_1 and box_2."""

    if box_1 is None and box_2 is None:
        return None
    if box_1 is None:
        return box_2.copy()
    if box_2 is None:
        return box_1.copy()
    result = np.zeros((2, 3), dtype = 'int')
    for dir in range(3):
        result[0][dir] = min(box_1[0][dir], box_2[0][dir])
        result[1][dir] = max(box_1[1][dir], box_2[1][dir])
    return result


def parent_cell_from_local_box_cell(box, box_cell, based_0_or_1 = 0):
    """Given a box and a local cell index triplet, converts to the equivalent cell index triplet in the host grid.

    input arguments (unmodified):
       box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          indices in the same ordering as box_cell; start value (python or Fortran/simulator) given by based_0_or_1
       box_cell: numpy int array of shape (3)
          indices of a cell within box, in coords local to box
          indices in the same ordering as box; start value (python or Fortran/simulator) given by based_0_or_1
       based_0_or_1: int, value 0 or 1
          start value (base) for indices of box and box_cell arguments, and of return value

    returns: numpy int array of shape (3)
       indices defining the cell in the host grid space equivalent to box_cell
       ordering of indices is same as that of box and box_cell; base is given by based_0_or_1 argument
    """

    assert box.ndim == 2 and box.shape == (2, 3)
    assert box_cell.ndim == 1 and box_cell.size == 3
    return box[0] + box_cell - based_0_or_1  # numpy 1D array op


def local_box_cell_from_parent_cell(box, parent_cell, based_0_or_1 = 0):
    """Given a cell index triplet in the host grid, and a box, returns the equivalent local cell index triplet.

    input arguments (unmodified):
       box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          indices in the same ordering as parent_cell; start value (python or Fortran/simulator) given by based_0_or_1
       parent_cell: numpy int array of shape (3)
          indices of a cell within host grid
          indices in the same ordering as box; start value (python or Fortran/simulator) given by based_0_or_1
       based_0_or_1: int, value 0 or 1
          start value (base) for indices of box and parent_cell arguments, and of return value

    returns: numpy int array of shape (3); or None
       indices defining the parent_cell in coords local to box, if the cell is within the box
       if parent_cell is not within box, None is returned
    """

    assert box.ndim == 2 and box.shape == (2, 3)
    assert parent_cell.ndim == 1 and parent_cell.size == 3
    if np.all(box[0] <= parent_cell) and np.all(parent_cell <= box[1]):  # numpy 1D array ops
        return parent_cell - box[0] + based_0_or_1  # numpy 1D array op
    else:
        return None


def boxes_overlap(box_a, box_b):
    """Returns True if the two boxes have any overlap in 3D, otherwise False.

    Arguments:
       box_a: numpy int or float array of shape (2, 3)
       box_b: numpy int or float array of shape (2, 3)

    if int arrays, each is lower & upper indices in 3 dimensions defining a logical cuboid
    subset of a 3D cartesian grid protocol of indices for the two boxes must be the same
    if float arrays, each is min & max x,y,z triplets

    returns: boolean
       True if box_a and box_b overlap, False otherwise
    """

    return not ((box_a[1, 0] < box_b[0, 0]) or (box_a[0, 0] > box_b[1, 0]) or (box_a[1, 1] < box_b[0, 1]) or
                (box_a[0, 1] > box_b[1, 1]) or (box_a[1, 2] < box_b[0, 2]) or (box_a[0, 2] > box_b[1, 2]))


def overlapping_boxes(established_box, new_box, trim_box):
    """Checks for 3D overlap of two boxes; returns True and sets trim_box if there is overlap, otherwise False.
    
    trim_box is modified in place.

    Arguments:
       established_box: numpy int array of shape (2, 3)
       new_box: numpy int array of shape (2, 3)
          each is lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          protocol of indices for the two boxes must be the same
       trim_box: numpy int array of shape (2, 3)
          set to lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          a subset of new_box such that if removed from new_box, a valid box would remain with no overlap with established_box
          indices protocol is the same as that used for established_box and new_box (if return value is True)
          if there is no overlap (return value False), all elements of trim_box are set to 0

    note:
       when there is overlap between the boxes, there can be more than one way to trim the new_box,
       with trim_box fully covering either ij, jk or ik planes of new_box
       the function selects the trim_box containing the minimum number of cells (minimum 'loss' to trimming)
       this function does not actually apply the trimming, ie. new_box is not modified here

    returns: boolean
       True if established_box and new_box overlap (implies trim_box valid), False otherwise (trim_box elements all 0)
    """

    assert established_box.ndim == 2 and established_box.shape == (2, 3) and established_box.dtype == 'int'
    assert new_box.ndim == 2 and new_box.shape == (2, 3) and new_box.dtype == 'int'
    assert trim_box.ndim == 2 and trim_box.shape == (2, 3) and trim_box.dtype == 'int'
    trim_box[:, :] = 0
    if not boxes_overlap(established_box, new_box):
        return False
    # determine trim direction based on minimizing number of cells to be trimmed
    new_box_area = np.zeros(
        3, dtype = 'int')  # compute number of cells in a 2D slice of new_box, taken in k, j or i directions
    # note: '_area' is a cell count, not an area in xyz space
    new_box_area[0] = (new_box[1, 1] - new_box[0, 1] + 1) * (new_box[1, 2] - new_box[0, 2] + 1)  # k slice
    new_box_area[1] = (new_box[1, 0] - new_box[0, 0] + 1) * (new_box[1, 2] - new_box[0, 2] + 1)  # j slice
    new_box_area[2] = (new_box[1, 0] - new_box[0, 0] + 1) * (new_box[1, 1] - new_box[0, 1] + 1)  # i slice
    # find the actual box of overlap and 'cost' of trimming in each dimension and minimize over k,j,i
    trim_cost = np.zeros(3, dtype = 'int')
    trim_direction = -1
    min_trim_cost = volume_of_box(new_box) + 1  # anything will be better than this
    overlap_box = np.zeros((2, 3), dtype = 'int')
    for dir in range(3):
        overlap_box[0, dir] = max(established_box[0, dir], new_box[0, dir])
        overlap_box[1, dir] = min(established_box[1, dir], new_box[1, dir])
        trim_cost[dir] = (overlap_box[1, dir] - overlap_box[0, dir] + 1) * new_box_area[dir]
        if trim_cost[dir] < min_trim_cost:
            min_trim_cost = trim_cost[dir]
            trim_direction = dir
    # set trim_box for chosen direction
    assert 0 <= trim_direction <= 2
    other_dir_a = (trim_direction + 1) % 3
    other_dir_b = (trim_direction + 2) % 3
    trim_box[0, trim_direction] = overlap_box[0, trim_direction]
    trim_box[1, trim_direction] = overlap_box[1, trim_direction]
    trim_box[0, other_dir_a] = new_box[0, other_dir_a]
    trim_box[1, other_dir_a] = new_box[1, other_dir_a]
    trim_box[0, other_dir_b] = new_box[0, other_dir_b]
    trim_box[1, other_dir_b] = new_box[1, other_dir_b]
    return True


def trim_box_by_box_returning_new_mask(box_to_be_trimmed, trim_box, mask_kji0):
    """Reduces box_to_be_trimmed by trim_box; trim_box must be a neat subset box at one face of box_to_be_trimmed.

    input/output argument (modified):
       box_to_be_trimmed: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          indices protocol is python kji ordering with zero base
          modified to exclude space occupied by trim_box

    input arguments (unmodified):
       trim_box: numpy int array of shape (2, 3)
          lower & upper indices in 3 dimensions defining a logical cuboid subset of a 3D cartesian grid
          indices protocol is python kji ordering with zero base
          the volume to be removed from box_to_be_trimmed,
          must be a subset of box_to_be_trimmed completely covering one face of box_to_be_trimmed
          (thereby ensuring that after trimming, a valid cuboid box results)
       mask_kji0: numpy 3D boolean array of shape matching extent of input box_to_be_trimmed
          indices protocol is python kji ordering with zero base

    returns: numpy 3D boolean array of shape matching extent of output box_to_be_trimmed
       the return array is a version of mask_kji0 that has been trimmed in accordance with the box trimming
    """

    local_box = np.zeros((2, 3), dtype = 'int')
    local_box[1] = extent_of_box(box_to_be_trimmed) - 1
    if np.all(trim_box[0] == box_to_be_trimmed[0]):  # shift box_to_be_trimmed [0] (minumum) up (in logical space)
        if trim_box[1, 0] != box_to_be_trimmed[1, 0]:  # trim in k direction
            assert (trim_box[1, 1] == box_to_be_trimmed[1, 1]) and (trim_box[1, 2] == box_to_be_trimmed[1, 2])
            local_box[0, 0] = trim_box[1, 0] - box_to_be_trimmed[0, 0] + 1
            box_to_be_trimmed[0, 0] = trim_box[1, 0] + 1
        elif trim_box[1, 1] != box_to_be_trimmed[1, 1]:  # trim in j direction
            assert (trim_box[1, 2] == box_to_be_trimmed[1, 2])
            local_box[0, 1] = trim_box[1, 1] - box_to_be_trimmed[0, 1] + 1
            box_to_be_trimmed[0, 1] = trim_box[1, 1] + 1
        else:  # trim in i direction
            assert trim_box[1, 2] != box_to_be_trimmed[1, 2]
            local_box[0, 2] = trim_box[1, 2] - box_to_be_trimmed[0, 2] + 1
            box_to_be_trimmed[0, 2] = trim_box[1, 2] + 1
    elif np.all(trim_box[1] == box_to_be_trimmed[1]):  # shift box_to_be_trimmed [1] (maxumum) down (in logical space)
        if trim_box[0, 0] != box_to_be_trimmed[0, 0]:  # trim in k direction
            assert (trim_box[0, 1] == box_to_be_trimmed[0, 1]) and (trim_box[0, 2] == box_to_be_trimmed[0, 2])
            local_box[1, 0] -= box_to_be_trimmed[1, 0] - trim_box[0, 0] + 1
            box_to_be_trimmed[1, 0] = trim_box[0, 0] - 1
        elif trim_box[0, 1] != box_to_be_trimmed[0, 1]:  # trim in j direction
            assert (trim_box[0, 2] == box_to_be_trimmed[0, 2])
            local_box[1, 1] -= box_to_be_trimmed[1, 1] - trim_box[0, 1] + 1
            box_to_be_trimmed[1, 1] = trim_box[0, 1] - 1
        else:  # trim in i direction
            assert trim_box[0, 2] != box_to_be_trimmed[0, 2]
            local_box[1, 2] -= box_to_be_trimmed[1, 2] - trim_box[0, 2] + 1
            box_to_be_trimmed[1, 2] = trim_box[0, 2] - 1
    else:  # shouldn't happen
        log.critical('Trim box code failure: box to be trimmed is %s', string_iijjkk1_for_box_kji0(box_to_be_trimmed))
        log.critical('Trim box code failure:          trim box is %s', string_iijjkk1_for_box_kji0(trim_box))
        assert False
    assert np.all(box_to_be_trimmed[0] <= box_to_be_trimmed[1])
    return (mask_kji0[local_box[0, 0]:local_box[1, 0] + 1, local_box[0, 1]:local_box[1, 1] + 1,
                      local_box[0, 2]:local_box[1, 2] + 1]).copy()


def trim_box_to_mask_returning_new_mask(bounding_box_kji0, mask_kji0):
    """Reduce the coverage of bounding box to the minimum needed to contain True elements of mask.
    
    Returns trimmed mask.
    """
    # NB: bounding box is modified by this function
    assert bounding_box_kji0.ndim == 2 and bounding_box_kji0.shape == (2, 3) and bounding_box_kji0.dtype == 'int'
    assert (mask_kji0.ndim == 3 and
            np.all(np.array([mask_kji0.shape], dtype = 'int') == extent_of_box(bounding_box_kji0)) and
            mask_kji0.dtype == 'bool')
    assert np.any(mask_kji0)
    box_extent = extent_of_box(bounding_box_kji0)
    local_box = np.zeros((2, 3), dtype = 'int')
    local_box[0] = box_extent.copy()
    local_box[1] = -1
    for k in range(box_extent[0]):
        if np.any(mask_kji0[k, :, :]):
            local_box[1, 0] = k
            if local_box[0, 0] == box_extent[0]:
                local_box[0, 0] = k
    for j in range(box_extent[1]):
        if np.any(mask_kji0[:, j, :]):
            local_box[1, 1] = j
            if local_box[0, 1] == box_extent[1]:
                local_box[0, 1] = j
    for i in range(box_extent[2]):
        if np.any(mask_kji0[:, :, i]):
            local_box[1, 2] = i
            if local_box[0, 2] == box_extent[2]:
                local_box[0, 2] = i
    assert np.all(local_box >= 0)
    assert np.all(local_box[1] >= local_box[0])
    assert np.all(local_box[1] < box_extent)
    bounding_box_kji0[0] += local_box[0]
    bounding_box_kji0[1] -= (box_extent - local_box[1] - 1)
    assert np.all(bounding_box_kji0[0] <= bounding_box_kji0[1])
    return (mask_kji0[local_box[0, 0]:local_box[1, 0] + 1, local_box[0, 1]:local_box[1, 1] + 1,
                      local_box[0, 2]:local_box[1, 2] + 1]).copy()


# end of box_utilities module



"""_stratigraphic_column_rank.py: RESQML StratigraphicColumnRankInterpretation class."""

version = '24th November 2021'

# NB: in this module, the term 'unit' refers to a geological stratigraphic unit, i.e. a layer of rock, not a unit of measure
# RMS is a registered trademark of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.strata._strata_common import valid_domains, _index_attr
from resqpy.strata._binary_contact_interpretation import BinaryContactInterpretation
from resqpy.strata._stratigraphic_unit_interpretation import StratigraphicUnitInterpretation


class StratigraphicColumnRank(BaseResqpy):
    """Class for RESQML StratigraphicColumnRankInterpretation objects.

    A list of stratigraphic unit interpretations, ordered from geologically oldest to youngest.
    """

    resqml_type = 'StratigraphicColumnRankInterpretation'

    # note: ordering of stratigraphic units and contacts is from geologically oldest to youngest

    def __init__(
            self,
            parent_model,
            uuid = None,
            domain = 'time',  # or should this be depth?
            rank_index = None,
            earth_model_feature_uuid = None,
            strata_uuid_list = None,  # ordered list of stratigraphic unit interpretations
            title = None,
            extra_metadata = None):
        """Initialises a Stratigraphic Column Rank resqpy object (RESQML StratigraphicColumnRankInterpretation).

        arguments:
           parent_model (model.Model): the model with which the new interpretation will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML stratigraphic column rank interpretation
              from which this object will be initialised
           domain (str, default 'time'): 'time', 'depth' or 'mixed', being the domain of the interpretation;
              ignored if uuid is not None
           rank_index (int, optional): the rank index (RESQML index) for this rank when multiple ranks are used;
              will default to zero if not set; ignored if uuid is not None
           earth_model_feature_uuid (uuid.UUID, optional): the uuid of an existing organization feature of kind
              'earth model' which this stratigraphic column is for; ignored if uuid is not None
           strata_uuid_list (list of uuid.UUID, optional): a list of uuids of existing stratigraphic unit
              interpretations, ordered from geologically oldest to youngest; ignored if uuid is not None
           title (str, optional): the citation title (feature name) of the new interpretation;
              ignored if uuid is not None
           extra_metadata (dict, optional): extra metadata items for the new stratigraphic column rank interpretation
        """

        self.feature_uuid = earth_model_feature_uuid  # interpreted earth model feature uuid
        self.domain = domain
        self.index = rank_index
        self.has_occurred_during = (None, None)  # optional RESQML item
        self.units = []  # ordered list of (index, stratagraphic unit interpretations)
        self.contacts = []  # optional ordered list of binary contact interpretations

        super().__init__(model = parent_model, uuid = uuid, title = title, extra_metadata = extra_metadata)

        if self.root is None and strata_uuid_list is not None:
            self.set_units(strata_uuid_list)

    def _load_from_xml(self):
        """Loads class specific attributes from xml for an existing RESQML object; called from BaseResqpy."""
        root_node = self.root
        assert root_node is not None
        assert rqet.find_tag_text(root_node, 'OrderingCriteria') == 'age',  \
           'stratigraphic column rank interpretation ordering criterion must be age'
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        self.feature_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(root_node, ['InterpretedFeature', 'UUID']))
        self.has_occurred_during = rqo.extract_has_occurred_during(root_node)
        self.index = rqet.find_tag_int(root_node, 'Index')
        self.units = []
        for su_node in rqet.list_of_tag(root_node, 'StratigraphicUnits'):
            index = rqet.find_tag_int(su_node, 'Index')
            unit_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(su_node, ['Unit', 'UUID']))
            assert index is not None and unit_uuid is not None
            assert self.model.type_of_uuid(unit_uuid, strip_obj = True) == 'StratigraphicUnitInterpretation'
            self.units.append((index, StratigraphicUnitInterpretation(self.model, uuid = unit_uuid)))
        self._sort_units()
        self.contacts = []
        for contact_node in rqet.list_of_tag(root_node, 'ContactInterpretation'):
            self.contacts.append(BinaryContactInterpretation(self.model, existing_xml_node = contact_node))
        self._sort_contacts()

    def set_units(self, strata_uuid_list):
        """Discard any previous units and set list of units based on ordered list of uuids.

        arguments:
           strata_uuid_list (list of uuid.UUID): the uuids of the stratigraphic unit interpretations which constitute
              this stratigraphic column, ordered from geologically oldest to youngest
        """

        self.units = []
        for i, uuid in enumerate(strata_uuid_list):
            assert self.model.type_of_uuid(uuid, strip_obj = True) == 'StratigraphicUnitInterpretation'
            self.units.append((i, StratigraphicUnitInterpretation(self.model, uuid = uuid)))

    def _sort_units(self):
        """Sorts units list, in situ, into increasing order of index values."""
        self.units.sort()

    def _sort_contacts(self):
        """Sorts contacts list, in situ, into increasing order of index values."""
        self.contacts.sort(key = _index_attr)

    def set_contacts_from_horizons(self, horizon_uuids, older_contact_mode = None, younger_contact_mode = None):
        """Sets the list of contacts from an ordered list of horizons, of length one less than the number of units.

        arguments:
           horizon_uuids (list of uuid.UUID): list of horizon interpretation uuids, ordered from geologically oldest
              to youngest, with one horizon for each neighbouring pair of stratigraphic units in this column
           older_contact_mode (str, optional): if present, the contact mode to set for the older unit for all of the
              contacts; must be in valid_contact_modes
           younger_contact_mode (str, optional): if present, the contact mode to set for the younger unit for all of
              the contacts; must be in valid_contact_modes

        notes:
           units must be established and sorted before calling this method; any previous contacts are discarded;
           if differing modes are required for the various contacts, leave the contact mode arguments as None,
           then interate over the contacts setting each mode individually (after this method returns)
        """

        # this method uses older unit as subject, younger as direct object
        # todo: check this is consistent with any RESQML usage guidance and/or any RMS usage

        self.contacts = []
        assert len(horizon_uuids) == len(self.units) - 1
        for i, horizon_uuid in enumerate(horizon_uuids):
            assert self.model.type_of_uuid(horizon_uuid, strip_obj = True) == 'HorizonInterpretation'
            contact = BinaryContactInterpretation(self.model,
                                                  index = i,
                                                  contact_relationship = 'stratigraphic unit to stratigraphic unit',
                                                  verb = 'stops at',
                                                  subject_uuid = self.units[i][1].uuid,
                                                  direct_object_uuid = self.units[i + 1][1].uuid,
                                                  subject_contact_side = 'older',
                                                  subject_contact_mode = older_contact_mode,
                                                  direct_object_contact_side = 'younger',
                                                  direct_object_contact_mode = younger_contact_mode,
                                                  part_of_uuid = horizon_uuid)
            self.contacts.append(contact)

    def iter_units(self):
        """Yields the ordered stratigraphic unit interpretations which constitute this stratigraphic colunn."""

        for _, unit in self.units:
            yield unit

    def unit_for_unit_index(self, index):
        """Returns the stratigraphic unit interpretation with the given index in this stratigraphic colunn."""

        for i, unit in self.units:
            if i == index:
                return unit
        return None

    def iter_contacts(self):
        """Yields the internal binary contact interpretations of this stratigraphic colunn."""

        for contact in self.contacts:
            yield contact

    # todo: implement a strict validation method checking contact exists for each neighbouring unit pair

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates a stratigraphic column rank interpretation xml tree.

        arguments:
           add_as_part (bool, default True): if True, the interpretation is added to the parent model as a high level part
           add_relationships (bool, default True): if True and add_as_part is True, a relationship is created with
              the referenced geologic unit feature
           originator (str, optional): if present, is used as the originator field of the citation block
           reuse (bool, default True): if True, the parent model is inspected for any equivalent interpretation and, if found,
              the uuid of this interpretation is set to that of the equivalent part

        returns:
           lxml.etree._Element: the root node of the newly created xml tree for the interpretation
        """

        # note: xml for referenced objects must be created before calling this method

        assert len(self.units), 'attempting to create xml for stratigraphic column rank without any units'

        if reuse and self.try_reuse():
            return self.root

        if self.index is None:
            self.index = 0

        scri = super().create_xml(add_as_part = False, originator = originator)

        assert self.feature_uuid is not None
        assert self.model.type_of_uuid(self.feature_uuid, strip_obj = True) == 'OrganizationFeature'
        # todo: could also check that the kind is 'earth model'
        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title(uuid = self.feature_uuid),
                                   self.feature_uuid,
                                   content_type = 'OrganizationFeature',
                                   root = scri)

        assert self.domain in valid_domains, 'illegal domain value for stratigraphic column rank interpretation'
        dom_node = rqet.SubElement(scri, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        rqo.create_xml_has_occurred_during(self.model, scri, self.has_occurred_during)

        oc_node = rqet.SubElement(scri, ns['resqml2'] + 'OrderingCriteria')
        oc_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'OrderingCriteria')
        oc_node.text = 'age'

        i_node = rqet.SubElement(scri, ns['resqml2'] + 'Index')
        i_node.set(ns['xsi'] + 'type', ns['xsi'] + 'nonNegativeInteger')
        i_node.text = str(self.index)

        for i, unit in self.units:

            su_node = rqet.SubElement(scri, ns['resqml2'] + 'StratigraphicUnits')
            su_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'StratigraphicUnitInterpretationIndex')
            su_node.text = ''

            si_node = rqet.SubElement(su_node, ns['resqml2'] + 'Index')
            si_node.set(ns['xsi'] + 'type', ns['xsi'] + 'nonNegativeInteger')
            si_node.text = str(i)

            self.model.create_ref_node('Unit',
                                       unit.title,
                                       unit.uuid,
                                       content_type = 'StratigraphicUnitInterpretation',
                                       root = su_node)

        for contact in self.contacts:
            contact.create_xml(scri)

        if add_as_part:
            self.model.add_part('obj_StratigraphicColumnRankInterpretation', self.uuid, scri)
            if add_relationships:
                emi_root = self.model.root(uuid = self.feature_uuid)
                self.model.create_reciprocal_relationship(scri, 'destinationObject', emi_root, 'sourceObject')
                for _, unit in self.units:
                    self.model.create_reciprocal_relationship(scri, 'destinationObject', unit.root, 'sourceObject')
                for contact in self.contacts:
                    if contact.part_of_uuid is not None:
                        horizon_root = self.model.root(uuid = contact.part_of_uuid)
                        if horizon_root is not None:
                            self.model.create_reciprocal_relationship(scri, 'destinationObject', horizon_root,
                                                                      'sourceObject')

        return scri



"""functions working with defined geometry in grids"""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.xml_et as rqet


def cell_geometry_is_defined(grid, cell_kji0 = None, cell_geometry_is_defined_root = None, cache_array = True):
    """Returns True if the geometry of the specified cell is defined.

    Can also be used to cache (load) the boolean array.

    arguments:
       cell_kji0 (triplet of integer, optional): if present, the index of the cell of interest, in kji0 protocol;
          if False, None is returned but the boolean array can still be cached
       cell_geometry_is_defined_root (optional): if present, the root of the 'cell geometry is defined' xml tree for
          this grid; this optional argument is to allow for speed optimisation, to save searching for the node
       cache_array (boolean, default True): if True, the 'cell geometry is defined' array is cached in memory, unless
          the xml tree indicates that geometry is defined for all cells, in which case that is noted

    returns:
       if cell_kji0 is not None, a boolean is returned indicating whether geometry is defined for that cell;
       if cell_kji0 is None, None is returned (but the array caching logic will have been applied)
    """

    if grid.geometry_defined_for_all_cells_cached:
        return True
    if hasattr(grid, 'array_cell_geometry_is_defined') and grid.array_cell_geometry_is_defined is None:
        delattr(grid, 'array_cell_geometry_is_defined')
    if hasattr(grid, 'array_cell_geometry_is_defined'):
        grid.geometry_defined_for_all_cells_cached = np.all(grid.array_cell_geometry_is_defined)
        if grid.geometry_defined_for_all_cells_cached:
            return True
        if cell_kji0 is None:
            return False
        return grid.array_cell_geometry_is_defined[tuple(cell_kji0)]
    is_def_root = grid.resolve_geometry_child('CellGeometryIsDefined', child_node = cell_geometry_is_defined_root)
    if is_def_root is None:
        points = grid.points_ref(masked = False)
        assert points is not None
        grid.geometry_defined_for_all_cells_cached = not np.any(np.isnan(points))
        if grid.geometry_defined_for_all_cells_cached or cell_kji0 is None:
            return grid.geometry_defined_for_all_cells_cached
    is_def_type = rqet.node_type(is_def_root)
    if is_def_type == 'BooleanConstantArray':
        grid.geometry_defined_for_all_cells_cached = (rqet.find_tag_text(is_def_root, 'Value').lower() == 'true')
        return grid.geometry_defined_for_all_cells_cached
    else:
        assert (is_def_type == 'BooleanHdf5Array')
        h5_key_pair = grid.model.h5_uuid_and_path_for_node(is_def_root)
        if h5_key_pair is None:
            return None
        result = grid.model.h5_array_element(h5_key_pair,
                                             index = cell_kji0,
                                             cache_array = cache_array,
                                             object = grid,
                                             array_attribute = 'array_cell_geometry_is_defined',
                                             dtype = 'bool')
        if grid.geometry_defined_for_all_cells_cached is None and cache_array and hasattr(
                grid, 'array_cell_geometry_is_defined'):
            grid.geometry_defined_for_all_cells_cached = (np.count_nonzero(
                grid.array_cell_geometry_is_defined) == grid.array_cell_geometry_is_defined.size)
            if grid.geometry_defined_for_all_cells_cached:
                delattr(grid, 'array_cell_geometry_is_defined')
        return result


def pillar_geometry_is_defined(grid, pillar_ji0 = None, pillar_geometry_is_defined_root = None, cache_array = True):
    """Returns True if the geometry of the specified pillar is defined; False otherwise.

    Can also be used to cache (load) the boolean array.

    arguments:
       pillar_ji0 (pair of integers, optional): if present, the index of the pillar of interest, in ji0 protocol;
          if False, None is returned but the boolean array can still be cached
       pillar_geometry_is_defined_root (optional): if present, the root of the 'pillar geometry is defined' xml tree for
          this grid; this optional argument is to allow for speed optimisation, to save searching for the node
       cache_array (boolean, default True): if True, the 'pillar geometry is defined' array is cached in memory, unless
          the xml tree indicates that geometry is defined for all pillars, in which case that is noted

    returns:
       if pillar_ji0 is not None, a boolean is returned indicating whether geometry is defined for that pillar;
       if pillar_ji0 is None, None is returned unless geometry is defined for all pillars in which case True is returned
    """

    if grid.geometry_defined_for_all_pillars_cached:
        return True
    if hasattr(grid, 'array_pillar_geometry_is_defined'):
        if pillar_ji0 is None:
            return None  # this option allows caching of array without actually referring to any pillar
        return grid.array_pillar_geometry_is_defined[tuple(pillar_ji0)]
    is_def_root = grid.resolve_geometry_child('PillarGeometryIsDefined', child_node = pillar_geometry_is_defined_root)
    if is_def_root is None:
        return True  # maybe default should be False?
    is_def_type = rqet.node_type(is_def_root)
    if is_def_type == 'BooleanConstantArray':
        assert rqet.find_tag(is_def_root, 'Value').text.lower() == 'true'
        grid.geometry_defined_for_all_pillars_cached = True
        return True
    else:
        assert is_def_type == 'BooleanHdf5Array'
        h5_key_pair = grid.model.h5_uuid_and_path_for_node(is_def_root)
        if h5_key_pair is None:
            return None
        result = grid.model.h5_array_element(h5_key_pair,
                                             index = pillar_ji0,
                                             cache_array = cache_array,
                                             object = grid,
                                             array_attribute = 'array_pillar_geometry_is_defined',
                                             dtype = 'bool')
        if grid.geometry_defined_for_all_pillars_cached is None and cache_array and hasattr(
                grid, 'array_pillar_geometry_is_defined'):
            grid.geometry_defined_for_all_pillars_cached = (np.count_nonzero(
                grid.array_pillar_geometry_is_defined) == grid.array_pillar_geometry_is_defined.size)
            if grid.geometry_defined_for_all_pillars_cached:
                del grid.array_pillar_geometry_is_defined  # memory optimisation
        return result


def geometry_defined_for_all_cells(grid, cache_array = True):
    """Returns True if geometry is defined for all cells; False otherwise.

    argument:
       cache_array (boolean, default True): if True, the 'cell geometry is defined' array is cached in memory,
          unless the xml indicates that geometry is defined for all cells, in which case that is noted

    returns:
       boolean: True if geometry is defined for all cells; False otherwise
    """

    if grid.geometry_defined_for_all_cells_cached is not None:
        return grid.geometry_defined_for_all_cells_cached
    if cache_array:
        grid.cell_geometry_is_defined(cache_array = True)
        return grid.geometry_defined_for_all_cells_cached
    # loop over all cells (until a False is encountered) – only executes if cache_array is False
    cell_geom_defined_root = grid.resolve_geometry_child('CellGeometryIsDefined')
    if cell_geom_defined_root is not None:
        for k0 in range(grid.nk):
            for j0 in range(grid.nj):
                for i0 in range(grid.ni):
                    if not grid.cell_geometry_is_defined(cell_kji0 = (k0, j0, i0),
                                                         cell_geometry_is_defined_root = cell_geom_defined_root,
                                                         cache_array = False):
                        grid.geometry_defined_for_all_cells_cached = False
                        return False
    grid.geometry_defined_for_all_cells_cached = True
    return True


def geometry_defined_for_all_pillars(grid, cache_array = True, pillar_geometry_is_defined_root = None):
    """Returns True if geometry is defined for all pillars; False otherwise.

    arguments:
       cache_array (boolean, default True): if True, the 'pillar geometry is defined' array is cached in memory,
          unless the xml indicates that geometry is defined for all pillars, in which case that is noted
       pillar_geometry_is_defined_root (optional): if present, the root of the 'pillar geometry is defined' xml tree for
          this grid; this optional argument is to allow for speed optimisation, to save searching for the node

    returns:
       boolean: True if the geometry is defined for all pillars; False otherwise
    """

    if grid.geometry_defined_for_all_pillars_cached is not None:
        return grid.geometry_defined_for_all_pillars_cached
    if cache_array:
        grid.pillar_geometry_is_defined(cache_array = cache_array)
        return grid.geometry_defined_for_all_pillars_cached
    is_def_root = grid.resolve_geometry_child('PillarGeometryIsDefined', child_node = pillar_geometry_is_defined_root)
    grid.geometry_defined_for_all_pillars_cached = True
    if is_def_root is not None:
        for pillar_j in range(grid.nj):
            for pillar_i in range(grid.ni):
                if not grid.pillar_geometry_is_defined(
                    [pillar_j, pillar_i], pillar_geometry_is_defined_root = is_def_root, cache_array = False):
                    grid.geometry_defined_for_all_pillars_cached = False
                    break
            if not grid.geometry_defined_for_all_pillars_cached:
                break
    return grid.geometry_defined_for_all_pillars_cached


def cell_geometry_is_defined_ref(grid):
    """Returns an in-memory numpy array containing the boolean data indicating which cells have geometry defined.

    returns:
       numpy array of booleans of shape (nk, nj, ni); True value indicates cell has geometry defined; False
       indicates that the cell's geometry (points xyz values) cannot be used

    note:
       if geometry is flagged in the xml as being defined for all cells, then this function returns None;
       geometry_defined_for_all_cells() can be used to test for that situation
    """

    # todo: treat this array like any other property?; handle constant array seamlessly?
    grid.cell_geometry_is_defined(cache_array = True)
    if hasattr(grid, 'array_cell_geometry_is_defined'):
        return grid.array_cell_geometry_is_defined
    return None  # can happen, if geometry is defined for all cells


def pillar_geometry_is_defined_ref(grid):
    """Returns an in-memory numpy array containing the boolean data indicating which pillars have geometry defined.

    returns:
       numpy array of booleans of shape (nj + 1, ni + 1); True value indicates pillar has geometry defined (at
       least for some points); False indicates that the pillar's geometry (points xyz values) cannot be used;
       the resulting array only covers primary pillars; extra pillars for split pillars always have geometry
       defined

    note:
       if geometry is flagged in the xml as being defined for all pillars, then this function returns None
    """

    # todo: double-check behaviour in presence of split pillars
    # todo: treat this array like any other property?; handle constant array seamlessly?
    grid.pillar_geometry_is_defined(cache_array = True)
    if hasattr(grid, 'array_pillar_geometry_is_defined'):
        return grid.array_pillar_geometry_is_defined
    return None  # can happen, if geometry is defined for all pillars


def set_geometry_is_defined(grid,
                            treat_as_nan = None,
                            treat_dots_as_nan = False,
                            complete_partial_pillars = False,
                            nullify_partial_pillars = False,
                            complete_all = False):
    """Set cached flags and/or arrays indicating which primary pillars have any points defined and which cells all points.

    arguments:
       treat_as_nan (float, optional): if present, any point with this value as x, y or z is changed
          to hold NaN values, which is the correct RESQML representation of undefined values
       treat_dots_as_nan (boolean, default False): if True, the points around any inactive cell which has zero length along
          all its I and J edges will be set to NaN (which can intentionally invalidate the geometry of neighbouring cells)
       complete_partial_pillars (boolean, default False): if True, pillars which have some but not all points defined will
          have values generated for the undefined (NaN) points
       nullify_partial_pillars (boolean, default False): if True, pillars which have some undefined (NaN) points will be
          treated as if all the points on the pillar are undefined
       complete_all (boolean, default False): if True, values will be generated for all undefined points (includes
          completion of partial pillars if both partial pillar arguments are False)

    notes:
       this method discards any previous information about which pillars and cells have geometry defined; the new settings
       are based solely on where points data is NaN (or has the value supplied as treat_as_nan etc.);
       the inactive attribute is also updated by this method, though any cells previously flagged as inactive will still be
       inactive;
       if points are generated due to either complete... argument being set True, the inactive mask is set prior to
       generating points, so all cells making use of generated points will be inactive; however, the geometry will show
       as defined where points have been generated;
       at most one of complete_partial_pillars and nullify_partial_pillars may be True;
       although the method modifies the cached (attribute) copies of various arrays, they are not written to hdf5 here
    """

    assert not (complete_partial_pillars and nullify_partial_pillars)
    if complete_all and not nullify_partial_pillars:
        complete_partial_pillars = True

    points = grid.points_ref(masked = False)

    if treat_as_nan is not None:
        nan_mask = np.any(np.logical_or(np.isnan(points), points == treat_as_nan), axis = -1)
    else:
        nan_mask = np.any(np.isnan(points), axis = -1)

    if treat_dots_as_nan:
        areal_dots = grid.point_areally()
        some_areal_dots = np.any(areal_dots)
    else:
        areal_dots = None
        some_areal_dots = False

    grid.geometry_defined_for_all_pillars_cached = None
    if hasattr(grid, 'array_pillar_geometry_is_defined'):
        del grid.array_pillar_geometry_is_defined
    grid.geometry_defined_for_all_cells_cached = None
    if hasattr(grid, 'array_cell_geometry_is_defined'):
        del grid.array_cell_geometry_is_defined

    if not np.any(nan_mask) and not some_areal_dots:
        grid.geometry_defined_for_all_pillars_cached = True
        grid.geometry_defined_for_all_cells_cached = True
        return

    if some_areal_dots:
        nan_mask = __handle_areal_dots(areal_dots, grid, nan_mask)

    assert not np.all(nan_mask), 'grid does not have any geometry defined'

    points[:] = np.where(np.repeat(np.expand_dims(nan_mask, axis = nan_mask.ndim), 3, axis = -1), np.NaN, points)

    surround_z = grid.xyz_box(lazy = False)[1 if grid.z_inc_down() else 0, 2]

    pillar_defined_mask = np.logical_not(np.all(nan_mask, axis = 0)).flatten()
    primary_count = (grid.nj + 1) * (grid.ni + 1)
    if np.all(pillar_defined_mask):
        grid.geometry_defined_for_all_pillars_cached = True
    else:
        grid.geometry_defined_for_all_pillars_cached = False
        grid.array_pillar_geometry_is_defined = pillar_defined_mask[:primary_count].reshape((grid.nj + 1, grid.ni + 1))
    if pillar_defined_mask.size > primary_count and not np.all(pillar_defined_mask[primary_count:]):
        log.warning('at least one split pillar has geometry undefined')

    grid.geometry_defined_for_all_cells_cached = False

    primary_nan_mask = \
        nan_mask.reshape((grid.nk_plus_k_gaps + 1, -1))[:, :primary_count].reshape(
            (grid.nk_plus_k_gaps + 1, grid.nj + 1, grid.ni + 1))
    column_nan_mask = np.logical_or(np.logical_or(primary_nan_mask[:, :-1, :-1], primary_nan_mask[:, :-1, 1:]),
                                    np.logical_or(primary_nan_mask[:, 1:, :-1], primary_nan_mask[:, 1:, 1:]))
    if grid.k_gaps:
        grid.array_cell_geometry_is_defined = np.logical_not(
            np.logical_or(column_nan_mask[grid.k_raw_index_array], column_nan_mask[grid.k_raw_index_array + 1]))
    else:
        grid.array_cell_geometry_is_defined = np.logical_not(np.logical_or(column_nan_mask[:-1], column_nan_mask[1:]))

    if hasattr(grid, 'inactive') and grid.inactive is not None:
        grid.inactive = np.logical_or(grid.inactive, np.logical_not(grid.array_cell_geometry_is_defined))
    else:
        grid.inactive = np.logical_not(grid.array_cell_geometry_is_defined)
    grid.all_inactive = np.all(grid.inactive)

    if grid.geometry_defined_for_all_cells_cached:
        return

    cells_update_needed = False

    if nullify_partial_pillars:
        partial_pillar_mask = np.logical_and(pillar_defined_mask, np.any(nan_mask, axis = 0).flatten())
        if np.any(partial_pillar_mask):
            points.reshape((grid.nk_plus_k_gaps + 1, -1, 3))[:, partial_pillar_mask, :] = np.NaN
            cells_update_needed = True
    elif complete_partial_pillars:
        partial_pillar_mask = np.logical_and(pillar_defined_mask, np.any(nan_mask, axis = 0).flatten())
        if np.any(partial_pillar_mask):
            log.warning('completing geometry for partially defined pillars')
            for pillar_index in np.where(partial_pillar_mask)[0]:
                __infill_partial_pillar(grid, pillar_index)
            cells_update_needed = True

    if complete_all:
        cells_update_needed = __complete_all_pillars(cells_update_needed, grid, points, surround_z)

    if cells_update_needed:
        # note: each pillar is either fully defined or fully undefined at this point
        if grid.geometry_defined_for_all_pillars_cached:
            grid.geometry_defined_for_all_cells_cached = True
            if hasattr(grid, 'array_cell_geometry_is_defined'):
                del grid.array_cell_geometry_is_defined
        else:
            top_nan_mask = np.isnan(points[0, ..., 0].flatten()[:(grid.nj + 1) * (grid.ni + 1)].reshape(
                (grid.nj + 1, grid.ni + 1)))
            column_nan_mask = np.logical_or(np.logical_or(top_nan_mask[:-1, :-1], top_nan_mask[:-1, 1:]),
                                            np.logical_or(top_nan_mask[1:, :-1], top_nan_mask[1:, 1:]))
            grid.array_cell_geometry_is_defined = np.repeat(np.expand_dims(column_nan_mask, 0), grid.nk, axis = 0)
            grid.geometry_defined_for_all_cells_cached = np.all(grid.array_cell_geometry_is_defined)
            if grid.geometry_defined_for_all_cells_cached:
                del grid.array_cell_geometry_is_defined


def __complete_all_pillars(cells_update_needed, grid, points, surround_z):
    # note: each pillar is either fully defined or fully undefined at this point
    top_nan_mask = np.isnan(points[0, ..., 0].flatten()[:(grid.nj + 1) * (grid.ni + 1)].reshape(
        (grid.nj + 1, grid.ni + 1)))
    surround_mask, coastal_mask = __create_surround_masks(top_nan_mask)
    holes_mask = np.logical_and(top_nan_mask, np.logical_not(surround_mask))
    if np.any(holes_mask):
        __fill_holes(grid, holes_mask)
    if np.any(surround_mask):
        __fill_surround(grid, surround_mask)
    # set z values for coastal and surround to max z for grid
    surround_mask = np.logical_or(surround_mask, coastal_mask).flatten()
    if np.any(surround_mask):
        points.reshape(grid.nk_plus_k_gaps + 1, -1, 3)[:, :(grid.nj + 1) * (grid.ni + 1)][:, surround_mask,
                                                                                          2] = surround_z
    grid.geometry_defined_for_all_pillars_cached = True
    if hasattr(grid, 'array_pillar_geometry_is_defined'):
        del grid.array_pillar_geometry_is_defined
    cells_update_needed = False
    assert not np.any(np.isnan(points))
    grid.geometry_defined_for_all_cells_cached = True
    if hasattr(grid, 'array_cell_geometry_is_defined'):
        del grid.array_cell_geometry_is_defined
    return cells_update_needed


def __handle_areal_dots(areal_dots, grid, nan_mask):
    # inject NaNs into the pillars around any cell that has zero length in I and J
    if grid.k_gaps:
        dot_mask = np.zeros((grid.nk_plus_k_gaps + 1, grid.nj + 1, grid.ni + 1), dtype = bool)
        dot_mask[grid.k_raw_index_array, :-1, :-1] = areal_dots
        dot_mask[grid.k_raw_index_array + 1, :-1, :-1] = np.logical_or(dot_mask[grid.k_raw_index_array + 1, :-1, :-1],
                                                                       areal_dots)
    else:
        dot_mask = np.zeros((grid.nk + 1, grid.nj + 1, grid.ni + 1), dtype = bool)
        dot_mask[:-1, :-1, :-1] = areal_dots
        dot_mask[1:, :-1, :-1] = np.logical_or(dot_mask[:-1, :-1, :-1], areal_dots)
    dot_mask[:, 1:, :-1] = np.logical_or(dot_mask[:, :-1, :-1], dot_mask[:, 1:, :-1])
    dot_mask[:, :, 1:] = np.logical_or(dot_mask[:, :, :-1], dot_mask[:, :, 1:])
    if grid.has_split_coordinate_lines:
        # only set points in primary pillars to NaN; todo: more thorough to consider split pillars too
        primaries = (grid.nj + 1) * (grid.ni + 1)
        nan_mask[:, :primaries] = np.logical_or(nan_mask[:, :primaries], dot_mask.reshape((-1, primaries)))
    else:
        nan_mask = np.where(dot_mask, np.NaN, nan_mask)
    return nan_mask


def __infill_partial_pillar(grid, pillar_index):
    points = grid.points_ref(masked = False).reshape((grid.nk_plus_k_gaps + 1, -1, 3))
    nan_mask = np.isnan(points[:, pillar_index, 0])
    first_k = 0
    while first_k < grid.nk_plus_k_gaps + 1 and nan_mask[first_k]:
        first_k += 1
    assert first_k < grid.nk_plus_k_gaps + 1
    if first_k > 0:
        points[:first_k, pillar_index] = points[first_k, pillar_index]
    last_k = grid.nk_plus_k_gaps
    while nan_mask[last_k]:
        last_k -= 1
    if last_k < grid.nk_plus_k_gaps:
        points[last_k + 1:, pillar_index] = points[last_k, pillar_index]
    while True:
        while first_k < last_k and not nan_mask[first_k]:
            first_k += 1
        if first_k >= last_k:
            break
        scan_k = first_k + 1
        while nan_mask[scan_k]:
            scan_k += 1
        points[first_k - 1: scan_k, pillar_index] = \
            np.linspace(points[first_k - 1, pillar_index], points[scan_k, pillar_index],
                        num=scan_k - first_k + 1, endpoint=False)
        first_k = scan_k


def __create_surround_masks(top_nan_mask):
    assert top_nan_mask.ndim == 2
    nj1, ni1 = top_nan_mask.shape  # nj + 1, ni + 1
    surround_mask = np.zeros(top_nan_mask.shape, dtype = bool)
    coastal_mask = np.zeros(top_nan_mask.shape, dtype = bool)
    for j in range(nj1):
        i = 0
        while i < ni1 and top_nan_mask[j, i]:
            coastal_mask[j, i] = True
            i += 1
        if i < ni1:
            i = ni1 - 1
            while top_nan_mask[j, i]:
                coastal_mask[j, i] = True
                i -= 1
        else:
            surround_mask[j] = True
            coastal_mask[j] = False
    for i in range(ni1):
        j = 0
        while j < nj1 and top_nan_mask[j, i]:
            coastal_mask[j, i] = True
            j += 1
        if j < nj1:
            j = nj1 - 1
            while top_nan_mask[j, i]:
                coastal_mask[j, i] = True
                j -= 1
        else:
            surround_mask[:, i] = True
            coastal_mask[:, i] = False
    return surround_mask, coastal_mask


def __fill_holes(grid, holes_mask):
    log.debug(f'filling {np.count_nonzero(holes_mask)} pillars for holes')
    points = grid.points_ref(masked = False).reshape(grid.nk_plus_k_gaps + 1, -1, 3)
    ni_plus_1 = grid.ni + 1
    mask_01 = np.empty(holes_mask.shape, dtype = int)
    while np.any(holes_mask):
        flat_holes_mask = holes_mask.flatten()
        mask_01[:] = np.where(holes_mask, 0, 1)
        modified = False
        # fix isolated NaN pillars with 4 neighbours
        neighbours = np.zeros(holes_mask.shape, dtype = int)
        neighbours[:-1, :] += mask_01[1:, :]
        neighbours[1:, :] += mask_01[:-1, :]
        neighbours[:, :-1] += mask_01[:, 1:]
        neighbours[:, 1:] += mask_01[:, :-1]
        foursomes = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 4))[0]
        if len(foursomes) > 0:
            interpolated = 0.25 * (points[:, foursomes - 1, :] + points[:, foursomes + 1, :] +
                                   points[:, foursomes - ni_plus_1, :] + points[:, foursomes + ni_plus_1, :])
            points[:, foursomes, :] = interpolated
            flat_holes_mask[foursomes] = False
            modified = True
        # fix NaN pillars with defined opposing neighbours in -J and +J
        neighbours[:] = 0
        neighbours[:-1, :] += mask_01[1:, :]
        neighbours[1:, :] += mask_01[:-1, :]
        twosomes = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 2))[0]
        if len(twosomes) > 0:
            interpolated = 0.5 * (points[:, twosomes - ni_plus_1, :] + points[:, twosomes + ni_plus_1, :])
            points[:, twosomes, :] = interpolated
            flat_holes_mask[twosomes] = False
            modified = True
        # fix NaN pillars with defined opposing neighbours in -I and +I
        neighbours[:] = 0
        neighbours[:, :-1] += mask_01[:, 1:]
        neighbours[:, 1:] += mask_01[:, :-1]
        twosomes = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 2))[0]
        if len(twosomes) > 0:
            interpolated = 0.5 * (points[:, twosomes - 1, :] + points[:, twosomes + 1, :])
            points[:, twosomes, :] = interpolated
            flat_holes_mask[twosomes] = False
            modified = True
        # fix NaN pillars with defined cornering neighbours in J- and I-
        neighbours[:] = 0
        neighbours[1:, :] += mask_01[:-1, :]
        neighbours[:, 1:] += mask_01[:, :-1]
        corners = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 2))[0]
        neighbours[1:, 1:] += mask_01[:-1, :-1]
        pushable = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 3))[0]
        if len(corners) > 0:
            interpolated = 0.5 * (points[:, corners - ni_plus_1, :] + points[:, corners - 1, :])
            points[:, corners, :] = interpolated
            pushed = 2.0 * points[:, pushable, :] - points[:, pushable - ni_plus_1 - 1, :]
            points[:, pushable, :] = pushed
            flat_holes_mask[corners] = False
            modified = True
        # fix NaN pillars with defined cornering neighbours in J- and I+
        neighbours[:] = 0
        neighbours[1:, :] += mask_01[:-1, :]
        neighbours[:, :-1] += mask_01[:, 1:]
        corners = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 2))[0]
        neighbours[1:, :-1] += mask_01[:-1, 1:]
        pushable = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 3))[0]
        if len(corners) > 0:
            interpolated = 0.5 * (points[:, corners - ni_plus_1, :] + points[:, corners + 1, :])
            points[:, corners, :] = interpolated
            pushed = 2.0 * points[:, pushable, :] - points[:, pushable - ni_plus_1 + 1, :]
            points[:, pushable, :] = pushed
            flat_holes_mask[corners] = False
            modified = True
        # fix NaN pillars with defined cornering neighbours in J+ and I-
        neighbours[:] = 0
        neighbours[:-1, :] += mask_01[1:, :]
        neighbours[:, 1:] += mask_01[:, :-1]
        corners = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 2))[0]
        neighbours[:-1, 1:] += mask_01[1:, :-1]
        pushable = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 3))[0]
        if len(corners) > 0:
            interpolated = 0.5 * (points[:, corners + ni_plus_1, :] + points[:, corners - 1, :])
            points[:, corners, :] = interpolated
            pushed = 2.0 * points[:, pushable, :] - points[:, pushable + ni_plus_1 - 1, :]
            points[:, pushable, :] = pushed
            flat_holes_mask[corners] = False
            modified = True
        # fix NaN pillars with defined cornering neighbours in J+ and I+
        neighbours[:] = 0
        neighbours[:-1, :] += mask_01[1:, :]
        neighbours[:, :-1] += mask_01[:, 1:]
        corners = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 2))[0]
        neighbours[:-1, :-1] += mask_01[1:, 1:]
        pushable = np.where(np.logical_and(flat_holes_mask, neighbours.flatten() == 3))[0]
        if len(corners) > 0:
            interpolated = 0.5 * (points[:, corners + ni_plus_1, :] + points[:, corners + 1, :])
            points[:, corners, :] = interpolated
            pushed = 2.0 * points[:, pushable, :] - points[:, pushable + ni_plus_1 + 1, :]
            points[:, pushable, :] = pushed
            flat_holes_mask[corners] = False
            modified = True
        holes_mask = flat_holes_mask.reshape((grid.nj + 1, ni_plus_1))
        if not modified:
            log.warning('failed to fill all holes in grid geometry')
            break


def __fill_surround(grid, surround_mask):
    # note: only fills x,y; based on bottom layer of points; assumes surround mask is a regularly shaped frame of columns
    log.debug(f'filling {np.count_nonzero(surround_mask)} pillars for surround')
    points = grid.points_ref(masked = False)
    points_view = points[-1, :, :2].reshape((-1, 2))[:(grid.nj + 1) * (grid.ni + 1), :].reshape(
        (grid.nj + 1, grid.ni + 1, 2))
    modified = False
    if grid.nj > 1:
        j_xy_vector = np.nanmean(points_view[1:, :] - points_view[:-1, :])
        j = 0
        while j < grid.nj and np.all(surround_mask[j, :]):
            j += 1
        assert j < grid.nj
        while j > 0:
            points_view[j - 1, :] = points_view[j, :] - j_xy_vector
            modified = True
            j -= 1
        j = grid.nj - 1
        while j >= 0 and np.all(surround_mask[j, :]):
            j -= 1
        assert j >= 0
        while j < grid.nj - 1:
            points_view[j + 1, :] = points_view[j, :] + j_xy_vector
            modified = True
            j += 1
    if grid.ni > 1:
        i_xy_vector = np.nanmean(points_view[:, 1:] - points_view[:, :-1])
        i = 0
        while i < grid.ni and np.all(surround_mask[:, i]):
            i += 1
        assert i < grid.ni
        while i > 0:
            points_view[:, i - 1] = points_view[:, i] - i_xy_vector
            modified = True
            i -= 1
        i = grid.ni - 1
        while i >= 0 and np.all(surround_mask[:, i]):
            i -= 1
        assert i >= 0
        while i < grid.ni - 1:
            points_view[:, i + 1] = points_view[:, i] + i_xy_vector
            modified = True
            i += 1
    if modified:
        points.reshape((grid.nk_plus_k_gaps + 1, -1, 3))[:-1, :(grid.nj + 1) * (grid.ni + 1), :2][:,
        surround_mask.flatten(), :] = \
            points_view[surround_mask, :].reshape((1, -1, 2))



"""High level add_wells_from_ascii_file() function."""

import logging

log = logging.getLogger(__name__)

import os

import resqpy.crs as rqc
import resqpy.model as rq
import resqpy.olio.xml_et as rqet
import resqpy.well as rqw

from resqpy.derived_model._common import _pl


def add_wells_from_ascii_file(epc_file,
                              crs_uuid,
                              trajectory_file,
                              comment_character = '#',
                              space_separated_instead_of_csv = False,
                              well_col = 'WELL',
                              md_col = 'MD',
                              x_col = 'X',
                              y_col = 'Y',
                              z_col = 'Z',
                              length_uom = 'm',
                              md_domain = None,
                              drilled = False,
                              z_inc_down = True,
                              new_epc_file = None):
    """Adds new md datum, trajectory, interpretation and feature objects for each well in a tabular ascii file..

    arguments:
       epc_file (string): file name to load model resqml model from (and rewrite to if new_epc_file is None)
       crs_uuid (uuid.UUID): the unique identifier of the coordinate reference system applicable to the x,y,z data;
          if None, a default crs will be created, making use of the length_uom and z_inc_down arguments
       trajectory_file (string): the path of the ascii file holding the well trajectory data to be loaded
       comment_character (string, default '#'): character deemed to introduce a comment in the trajectory file
       space_separated_instead_of_csv (boolean, default False): if True, the columns in the trajectory file are space
          separated; if False, comma separated
       well_col (string, default 'WELL'): the heading for the column containing well names
       md_col (string, default 'MD'): the heading for the column containing measured depths
       x_col (string, default 'X'): the heading for the column containing X (usually easting) data
       y_col (string, default 'Y'): the heading for the column containing Y (usually northing) data
       z_col (string, default 'Z'): the heading for the column containing Z (depth or elevation) data
       length_uom (string, default 'm'): the units of measure for the measured depths; should be 'm' or 'ft'
       md_domain (string, optional): the source of the original deviation data; may be 'logger' or 'driller'
       drilled (boolean, default False): True should be used for wells that have been drilled; False otherwise (planned,
          proposed, or a location being studied)
       z_inc_down (boolean, default True): indicates whether z values increase with depth; only used in the creation
          of a default coordinate reference system; ignored if crs_uuid is not None
       new_epc_file (string, optional): if None, the source epc_file is extended with the new property object; if present,
          a new epc file (& associated h5 file) is created to contain a copy of the grid and the new property

    returns:
       int: the number of wells added

    notes:
       ascii file must be table with first line being column headers, with columns for WELL, MD, X, Y & Z;
       actual column names can be set with optional arguments;
       all the objects are added to the model, with array data being written to the hdf5 file for the trajectories;
       the md_domain and drilled values are stored in the RESQML metadata but are only for human information and do not
       generally affect computations
    """

    assert trajectory_file and os.path.exists(trajectory_file)
    if md_domain:
        assert md_domain in ['driller', 'logger']

    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None

    # open up model
    if new_epc_file:
        model = rq.Model(new_epc_file, copy_from = epc_file)
    else:
        model = rq.Model(epc_file)

    # sort out the coordinate reference system
    if crs_uuid is None:
        crs_uuid = model.crs_uuid
    if crs_uuid is None:
        if z_inc_down is None:
            z_inc_down = True
        crs = rqc.Crs(model, xy_units = length_uom, z_units = length_uom, z_inc_down = z_inc_down)
        crs.create_xml()
        crs_uuid = crs.uuid

    # add all the well related objects to the model, based on data in the ascii file
    (feature_list, interpretation_list, trajectory_list, md_datum_list) =  \
       rqw.add_wells_from_ascii_file(model, crs_uuid, trajectory_file, comment_character = comment_character,
                                     space_separated_instead_of_csv = space_separated_instead_of_csv,
                                     well_col = well_col, md_col = md_col, x_col = x_col, y_col = y_col, z_col = z_col,
                                     length_uom = length_uom, md_domain = md_domain, drilled = drilled)

    assert len(feature_list) == len(interpretation_list) == len(trajectory_list) == len(md_datum_list)
    count = len(feature_list)

    log.info('features, interpretations, trajectories and md data added for ' + str(count) + ' well' + _pl(count))

    # write or re-write model
    model.h5_release()
    model.store_epc()

    return count



"""Multiprocessing module for running specific functions concurrently."""

__all__ = ['function_multiprocessing', 'find_faces_to_represent_surface_regular_wrapper']

from .multiprocessing import function_multiprocessing
from .wrappers.grid_surface import find_faces_to_represent_surface_regular_wrapper

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""A function for writing out a Nexus CORP file using supplied grid geometry"""

import logging

import numpy as np

log = logging.getLogger(__name__)

import resqpy.olio.grid_functions as gf
import resqpy.olio.xml_et as rqet
import resqpy.olio.trademark as tm
import resqpy.olio.write_data as wd
import resqpy.crs as rqc


def write_nexus_corp(grid,
                     file_name,
                     local_coords = False,
                     global_xy_units = None,
                     global_z_units = None,
                     global_z_increasing_downward = True,
                     write_nx_ny_nz = False,
                     write_units_keyword = False,
                     write_rh_keyword_if_needed = False,
                     write_corp_keyword = False,
                     use_binary = False,
                     binary_only = False,
                     nan_substitute_value = None):
    """Write grid geometry to file in Nexus CORP ordering."""

    log.info('caching Nexus corner points')
    tm.log_nexus_tm('info')
    grid.corner_points(cache_cp_array = True)
    log.debug('duplicating Nexus corner points')
    cp = grid.array_corner_points.copy()
    log.debug('resequencing duplicated Nexus corner points')
    gf.resequence_nexus_corp(cp, eight_mode = False, undo = True)
    corp_extent = np.zeros(3, dtype = 'int')
    corp_extent[0] = grid.cell_count()  # total number of cells in grid
    corp_extent[1] = 8  # 8 corners of cell: k -/+; j -/+; i -/+
    corp_extent[2] = 3  # x, y, z
    ijk_right_handed = grid.extract_grid_is_right_handed()
    if ijk_right_handed is None:
        log.warning('ijk handedness not known')
    elif not ijk_right_handed:
        log.warning('ijk axes are left handed; inverted (fake) xyz handedness required')
    crs_uuid = grid.crs_uuid
    if not local_coords:
        if not global_z_increasing_downward:
            log.warning('global z is not increasing with depth as expected by Nexus')
            tm.log_nexus_tm('warning')
        if crs_uuid is not None:  # todo: otherwise raise exception?
            log.info('converting corner points from local to global reference system')
            cp = grid.local_to_global_crs(cp,
                                          crs_uuid = crs_uuid,
                                          global_xy_units = global_xy_units,
                                          global_z_units = global_z_units,
                                          global_z_increasing_downward = global_z_increasing_downward)
    log.info('writing simulator corner point file ' + file_name)
    with open(file_name, 'w') as header:
        header.write('! Nexus corner point data written by resqml_grid module\n')
        header.write('! Nexus is a registered trademark of the Halliburton Company\n\n')
        crs = rqc.Crs(grid.model, uuid = crs_uuid)
        if write_units_keyword:
            if local_coords:
                if crs_uuid is not None:
                    if crs.xy_units == 'm' and crs.z_units == 'm':
                        header.write('METRIC\n\n')
                    elif crs.xy_units.startswith('ft') and crs.z_units.startswith('ft'):
                        header.write('ENGLISH\n\n')
                    else:
                        header.write('! local coordinates mixed (or not recognized)\n\n')
                else:
                    header.write('! local coordinates unknown\n\n')
            elif global_xy_units is not None and global_z_units is not None and global_xy_units == global_z_units:
                if global_xy_units in ['m', 'metre', 'metres']:
                    header.write('METRIC\n\n')
                elif global_xy_units in ['ft', 'feet', 'foot']:
                    header.write('ENGLISH\n\n')
                else:
                    header.write('! global coordinates not recognized\n\n')
            else:
                header.write('! global units unknown or mixed\n\n')
        if write_nx_ny_nz:
            header.write('NX      NY      NZ\n')
            header.write('{0:<7d} {1:<7d} {2:<7d}\n\n'.format(grid.extent_kji[2], grid.extent_kji[1],
                                                              grid.extent_kji[0]))
        if write_rh_keyword_if_needed:
            if ijk_right_handed is None or crs_uuid is None:
                log.warning('unable to determine whether RIGHTHANDED keyword is needed')
            else:
                if local_coords:
                    z_inc_down = grid.z_inc_down()
                    if not z_inc_down:
                        log.warning('local z is not increasing with depth as expected by Nexus')
                        tm.log_nexus_tm('warning')
                else:
                    z_inc_down = global_z_increasing_downward
                xyz_handedness = rqet.xyz_handedness(crs.axis_order, z_inc_down)
                if xyz_handedness == 'unknown':
                    log.warning(
                        'xyz handedness is not known; unable to determine whether RIGHTHANDED keyword is needed')
                else:
                    if ijk_right_handed == (xyz_handedness == 'right'):  # if either both True or both False
                        header.write('RIGHTHANDED\n\n')
    if write_corp_keyword:
        keyword = 'CORP VALUE'
    else:
        keyword = None
    wd.write_array_to_ascii_file(file_name,
                                 corp_extent,
                                 cp.reshape(tuple(corp_extent)),
                                 target_simulator = 'nexus',
                                 keyword = keyword,
                                 columns = 3,
                                 blank_line_after_i_block = False,
                                 blank_line_after_j_block = True,
                                 append = True,
                                 use_binary = use_binary,
                                 binary_only = binary_only,
                                 nan_substitute_value = nan_substitute_value)



"""Classes for RESQML objects related to surfaces."""

__all__ = [
    'BaseSurface', 'CombinedSurface', 'Mesh', 'TriangulatedPatch', 'PointSet', 'Surface', 'distill_triangle_points'
]

from ._base_surface import BaseSurface
from ._combined_surface import CombinedSurface
from ._mesh import Mesh
from ._triangulated_patch import TriangulatedPatch
from ._pointset import PointSet
from ._surface import Surface, distill_triangle_points

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Module providing wrapper for random number generator seeding functions."""

version = '6th November 2020'

import random

import numpy


def seed(seed, package = 'all'):
    """Set seed for random number generator of one or more packages, to allow for repeatable behaviour.

    arguments:
       seed (int or long): the value to use to seed the random number generator(s); a value of None will
          generally result in an unrepeatable sequence
       package (string or list of strings): one or more of known packages: 'random' and 'numpy' at present;
          passing 'all' will cause all packages known to have a random number generator to be re-seeded
    """

    known_list = ['random', 'numpy']

    if isinstance(package, str):
        if package == 'all':
            package = known_list
        else:
            package = [package]

    assert isinstance(package, list)

    for pack in package:
        assert pack in known_list, 'unknown package for random number seeding: ' + str(pack)

        if pack == 'random':
            random.seed(seed)
        elif pack == 'numpy':
            numpy.random.seed(seed = seed)



"""_pointset.py: pointset class based on resqml standard."""

version = '18th February 2022'

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company
# GOCAD is also a trademark of Emerson

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.crs as rcrs
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
from resqpy.olio.xml_namespaces import curly_namespace as ns
from ._base_surface import BaseSurface


class PointSet(BaseSurface):
    """Class for RESQML Point Set Representation within resqpy model object."""  # TODO: Work in Progress

    resqml_type = 'PointSetRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 load_hdf5 = False,
                 points_array = None,
                 crs_uuid = None,
                 polyset = None,
                 polyline = None,
                 random_point_count = None,
                 charisma_file = None,
                 irap_file = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Creates an empty Point Set object and optionally populates from xml or other source.

        arguments:
           parent_model (model.Model object): the model to which the new point set belongs
           uuid (uuid.UUID, optional): if present, the object is populated from the RESQML PointSetRepresentation
              with this uuid
           load_hdf5 (boolean, default False): if True and uuid is present, the actual points are
              pre-loaded into a numpy array; otherwise the points will be loaded on demand
           points_array (numpy float array of shape (..., 2 or 3), optional): if present, the xy(&z) data which
              will constitute the point set; missing z will be set to zero; ignored if uuid is not None
           crs_uuid (uuid.UUID, optional): if present, identifies the coordinate reference system for the points;
              ignored if uuid is not None; if None, 'imported' points will be associated with the
              default crs of the parent model
           polyset (optional): if present, creates a pointset from points in a polylineset
           polyline (optional): if present and random_point_count is None or zero, creates a pointset from
              points in a polyline; if present and random_point_count is set, creates random points within
              the (closed, convex) polyline
           random_point_count (int, optional): if present and polyline is present then the number of random
              points to generate within the (closed) polyline in the xy plane, with z set to 0.0
           charisma_file (optional): if present, creates a pointset from a charisma 3d interpretation file
           irap_file (optional): if present, creates a pointset from an IRAP classic points format file
           title (str, optional): the citation title to use for a new point set;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the point set, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the point set;
              ignored if uuid is not None

        returns:
           newly created PointSet object

        :meta common:
        """

        self.crs_uuid = crs_uuid
        self.patch_count = None
        self.patch_ref_list = []  # ordered list of (patch hdf5 ext uuid, path in hdf5, point count)
        self.patch_array_list = []  # ordered list of numpy float arrays (or None before loading), each of shape (N, 3)
        self.full_array = None  # composite points (all patches)
        self.represented_interpretation_root = None
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            if load_hdf5:
                self.load_all_patches()

        elif points_array is not None:
            assert self.crs_uuid is not None, 'missing crs uuid when establishing point set from array'
            self.add_patch(points_array)

        elif polyline is not None:  # Points from or within polyline
            self.from_polyline(polyline, random_point_count)

        elif polyset is not None:  # Points from polylineSet
            self.from_polyset(polyset)

        elif charisma_file is not None:  # Points from Charisma 3D interpretation lines
            self.from_charisma(charisma_file)

        elif irap_file is not None:  # Points from IRAP simple points
            self.from_irap(irap_file)

        if not self.title:
            self.title = 'point set'

    def _load_from_xml(self):
        assert self.root is not None
        self.patch_count = rqet.count_tag(self.root, 'NodePatch')
        assert self.patch_count, 'no patches found in xml for point set'
        self.patch_array_list = [None for _ in range(self.patch_count)]
        patch_index = 0
        for child in rqet.list_of_tag(self.root, 'NodePatch'):
            point_count = rqet.find_tag_int(child, 'Count')
            geom_node = rqet.find_tag(child, 'Geometry')
            assert geom_node is not None, 'geometry missing in xml for point set patch'
            crs_uuid = rqet.find_nested_tags_text(geom_node, ['LocalCrs', 'UUID'])
            assert crs_uuid, 'crs uuid missing in geometry xml for point set patch'
            self.check_crs_match(crs_uuid)
            ext_uuid = rqet.find_nested_tags_text(geom_node, ['Points', 'Coordinates', 'HdfProxy', 'UUID'])
            assert ext_uuid, 'missing hdf5 uuid in geometry xml for point set patch'
            hdf5_path = rqet.find_nested_tags_text(geom_node, ['Points', 'Coordinates', 'PathInHdfFile'])
            assert hdf5_path, 'missing internal hdf5 path in geometry xml for point set patch'
            self.patch_ref_list.append((ext_uuid, hdf5_path, point_count))
            patch_index += 1

        ref_node = rqet.find_tag(self.root, 'RepresentedInterpretation')
        if ref_node is not None:
            interp_root = self.model.referenced_node(ref_node)
            self.set_represented_interpretation_root(interp_root)
        # note: load of patches handled elsewhere

    def from_charisma(self, charisma_file):
        """Instantiate a pointset using points from an input charisma file

        arguments:
            charisma_file: a charisma 3d interpretation file
        """
        with open(charisma_file, 'r') as surf:
            for i, line in enumerate(surf.readlines()):
                if i == 0:
                    cpoints = np.array([[float(x) for x in line.split()[6:]]])
                else:
                    curr = np.array([[float(x) for x in line.split()[6:]]])
                    cpoints = np.concatenate((cpoints, curr))
        self.add_patch(cpoints)
        assert self.crs_uuid is not None, 'crs uuid missing when establishing point set from charisma file'
        if not self.title:
            self.title = charisma_file

    def from_irap(self, irap_file):
        """Instantiate a pointset using points from an input irap file.

        arguments:
            irap_file: a IRAP classic points format file
        """
        with open(irap_file, 'r') as points:
            for i, line in enumerate(points.readlines()):
                if i == 0:
                    cpoints = np.array([[float(x) for x in line.split(" ")]])
                else:
                    curr = np.array([[float(x) for x in line.split(" ")]])
                    cpoints = np.concatenate((cpoints, curr))
        self.add_patch(cpoints)
        assert self.crs_uuid is not None, 'crs uuid missing when establishing point set from irap file'
        if not self.title:
            self.title = irap_file

    def from_polyline(self, polyline, random_point_count = None):
        """Instantiates a pointset using points from an input polyline (PolylineRepresentation) object

        arguments:
            polyline (resqpy.lines.Polyline object): if random_point_count is None or zero, creates a pointset from
              points in a polyline; if present and random_point_count is set, creates random points within
              the (closed, convex) polyline
           random_point_count (int, optional): if present then the number of random
              points to generate within the (closed) polyline in the xy plane, with z set to 0.0

        """
        if random_point_count:
            assert polyline.is_convex()
            points = np.zeros((random_point_count, 3))
            points[:, :2] = np.random.random((random_point_count, 2))
            for p_i in range(random_point_count):
                points[p_i, :2] = polyline.denormalised_xy(points[p_i, 0], points[p_i, 1])
            self.add_patch(points)
        else:
            self.add_patch(polyline.coordinates)
            if polyline.rep_int_root is not None:
                self.set_represented_interpretation_root(polyline.rep_int_root)
        self.check_crs_match(polyline.crs_uuid)
        if not self.title:
            self.title = polyline.title

    def from_polyset(self, polyset):
        """Instantiates a pointset using points from an input polylineset (PolylineSetRepresentation) object

        arguments:
            polyset (resqpy.lines.PolylineSet object): a polylineset object to generate the pointset from
        """
        master_crs = rcrs.Crs(self.model, uuid = polyset.polys[0].crs_uuid)
        if polyset.polys[0].isclosed and vec.isclose(polyset.polys[0].coordinates[0], polyset.polys[0].coordinates[-1]):
            poly_coords = polyset.polys[0].coordinates[:-1].copy()
        else:
            poly_coords = polyset.polys[0].coordinates.copy()
        for poly in polyset.polys[1:]:
            curr_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
            assert master_crs is not None
            assert poly_coords is not None
            if not curr_crs.is_equivalent(master_crs):
                shifted = curr_crs.convert_array_to(master_crs, poly.coordinates)
                poly_coords = concat_polyset_points(poly.isclosed, shifted, poly_coords)
            else:
                poly_coords = concat_polyset_points(poly.isclosed, poly.coordinates, poly_coords)
        self.add_patch(poly_coords)
        if polyset.rep_int_root is not None:
            self.set_represented_interpretation_root(polyset.rep_int_root)
        self.check_crs_match(master_crs.uuid)
        if not self.title:
            self.title = polyset.title

    def check_crs_match(self, crs_uuid):
        """Checks if the model crs_uuid matches a given crs_uuid. If the current model crs_uuid is None, the new crs_uuid is used as the model crs_uuid. If a mismatch is found an assertion error is raised."""
        if self.crs_uuid is None:
            self.crs_uuid = crs_uuid
        else:
            assert bu.matching_uuids(crs_uuid, self.crs_uuid), 'mixed coordinate reference systems in point set'

    def set_represented_interpretation_root(self, interp_root):
        """Makes a note of the xml root of the represented interpretation."""

        self.represented_interpretation_root = interp_root

    def single_patch_array_ref(self, patch_index):
        """Load numpy array for one patch of the point set from hdf5, cache and return it."""

        assert 0 <= patch_index < self.patch_count, 'point set patch index out of range'
        if self.patch_array_list[patch_index] is not None:
            return self.patch_array_list[patch_index]
        h5_key_pair = (self.patch_ref_list[patch_index][0], self.patch_ref_list[patch_index][1]
                      )  # ext uuid, path in hdf5
        try:
            self.model.h5_array_element(h5_key_pair,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'temp_points',
                                        dtype = 'float')
        except Exception:
            log.exception('hdf5 points failure for point set patch ' + str(patch_index))
        assert self.temp_points.ndim == 2 and self.temp_points.shape[
            1] == 3, 'unsupported dimensionality to points array'
        self.patch_array_list[patch_index] = self.temp_points.copy()
        delattr(self, 'temp_points')
        return self.patch_array_list[patch_index]

    def load_all_patches(self):
        """Load hdf5 data for all patches and cache as separate numpy arrays; not usually called directly."""

        for patch_index in range(self.patch_count):
            self.single_patch_array_ref(patch_index)

    def change_crs(self, required_crs):
        """Changes the crs of the point set, also sets a new uuid if crs changed.

        notes:
           this method is usually used to change the coordinate system for a temporary resqpy object;
           to add as a new part, call write_hdf5() and create_xml() methods
        """

        old_crs = rcrs.Crs(self.model, uuid = self.crs_uuid)
        self.crs_uuid = required_crs.uuid
        if required_crs == old_crs or not self.patch_ref_list:
            log.debug(f'no crs change needed for {self.title}')
            return
        log.debug(f'crs change needed for {self.title} from {old_crs.title} to {required_crs.title}')
        self.load_all_patches()
        self.patch_ref_list = []
        for patch_points in self.patch_array_list:
            required_crs.convert_array_from(old_crs, patch_points)
            self.patch_ref_list.append((None, None, len(patch_points)))
        self.full_array = None  # clear cached full array for point set
        self.uuid = bu.new_uuid()  # hope this doesn't cause problems

    def trim_to_xyz_box(self, xyz_box):
        """Converts point set to a single patch, holding only those points within the xyz box.

        arguments:
           xyz_box (numpy float array of shape (2, 3)): the minimum and maximum range to keep in x,y,z

        notes:
           usually used to reduce the point set volume for a temprary object; a new uuid is assigned;
           to add as a new part, call write_hdf5() and create_xml() methods
        """
        points = self.full_array_ref()
        keep_mask = np.where(
            np.logical_and(np.all(points >= np.expand_dims(xyz_box[0], axis = 0), axis = -1),
                           np.all(points <= np.expand_dims(xyz_box[1], axis = 0), axis = -1)))
        self.patch_count = 0
        self.patch_ref_list = []
        self.patch_array_list = []
        self.full_array = None
        self.add_patch(points[keep_mask, :].copy())
        self.uuid = bu.new_uuid()  # hope this doesn't cause problems

    def trim_to_xy_polygon(self, xy_polygon):
        """Converts point set to a single patch, holding only those points within the polygon when projected in xy.

        arguments:
           xy_polygon (closed convex resqpy.lines.Polyline): the polygon outlining the area in xy within which
              points are kept

        notes:
           usually used to reduce the point set volume for a temprary object; a new uuid is assigned;
           to add as a new part, call write_hdf5() and create_xml() methods
        """
        points = self.full_array_ref()
        keep_mask = xy_polygon.points_are_inside_xy(points)
        self.patch_count = 0
        self.patch_ref_list = []
        self.patch_array_list = []
        self.full_array = None
        self.add_patch(points[keep_mask, :].copy())
        self.uuid = bu.new_uuid()  # hope this doesn't cause problems

    def minimum_xy_area_rectangle(self, delta_theta = 5.0):
        """Returns the xy projection rectangle of minimum area that contains the points.

        arguments:
           delta_theta (float, default 5.0): the incremental angle in degrees to test against

        returns:
           (d1, d2, r) where d1 and d2 are lengths of sides of an xy plane rectangle that just contains the
           points, and d1 <= d2, and r is a bearing in degrees of a d2 (longer) side between 0.0 and 180.0
        """

        def try_angle(pset, theta):
            p = pset.full_array_ref().copy()
            m = vec.rotation_matrix_3d_axial(0, theta)
            p = vec.rotate_array(m, p)[:, :2]
            dxy = np.nanmax(p, axis = 0) - np.nanmin(p, axis = 0)
            return dxy

        assert self.patch_count > 0
        theta = 0.0
        min_dxy = try_angle(self, theta)
        min_area = np.sum(min_dxy * min_dxy)
        while theta < 180.0 - delta_theta:
            theta += delta_theta
            dxy = try_angle(self, theta)
            area = np.sum(dxy * dxy)
            if area < min_area:
                min_dxy = dxy
        if dxy[0] <= dxy[1]:
            return dxy[0], dxy[1], theta
        return dxy[1], dxy[0], theta + 90.0 if theta < 90.0 else theta - 90.0

    def full_array_ref(self):
        """Return a single numpy float array of shape (N, 3) containing all points from all patches.

        :meta common:
        """

        if self.full_array is not None:
            return self.full_array
        self.load_all_patches()
        if self.patch_count == 1:  # optimisation, as usually the case
            self.full_array = self.patch_array_list[0]
            return self.full_array
        point_count = 0
        for patch_index in range(self.patch_count):
            point_count += self.patch_ref_list[patch_index][2]
        self.full_array = np.empty((point_count, 3))
        full_index = 0
        for patch_index in range(self.patch_count):
            self.full_array[full_index:full_index +
                            self.patch_ref_list[patch_index][2]] = self.patch_array_list[patch_index]
            full_index += self.patch_ref_list[patch_index][2]
        assert full_index == point_count, 'point count mismatch when constructing full array for point set'
        return self.full_array

    def add_patch(self, points_array):
        """Extend the current point set with a new patch of points."""

        assert points_array.ndim >= 2 and points_array.shape[-1] in [2, 3]
        if points_array.shape[-1] == 2:
            shape = list(points_array.shape)
            shape[-1] = 3
            p = np.zeros(shape)
            p[..., :2] = points_array
            points_array = p
        self.patch_array_list.append(points_array.reshape(-1, 3).copy())
        self.patch_ref_list.append((None, None, points_array.shape[0]))
        self.full_array = None
        if self.patch_count is None:
            self.patch_count = 0
        self.patch_count += 1

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Create or append to an hdf5 file, writing datasets for the point set patches after caching arrays.

        :meta common:
        """

        if not file_name:
            file_name = self.model.h5_file_name()
        if self.uuid is None:
            self.uuid = bu.new_uuid()
        # NB: patch arrays must all have been set up prior to calling this function
        h5_reg = rwh5.H5Register(self.model)
        for patch_index in range(self.patch_count):
            h5_reg.register_dataset(self.uuid, 'points_{}'.format(patch_index), self.patch_array_list[patch_index])
        h5_reg.write(file_name, mode = mode)

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None):
        """Creates a point set representation xml node from this point set object and optionally adds as part of model.

        arguments:
            ext_uuid (uuid.UUID): the uuid of the hdf5 external part holding the points array(s)
            add_as_part (boolean, default True): if True, the newly created xml node is added as a part
                in the model
            add_relationships (boolean, default True): if True, a relationship xml part is created relating the
                new point set part to the crs part (and optional interpretation part)
            title (string): used as the citation Title text; should be meaningful to a human
            originator (string, optional): the name of the human being who created the point set representation part;
                default is to use the login name

        returns:
            the newly created point set representation xml node

        :meta common:
        """

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        ps_node = super().create_xml(add_as_part = False, title = title, originator = originator)

        if self.crs_uuid is None:
            self.crs_uuid = self.model.crs_uuid  # maverick use of model's default crs

        if self.represented_interpretation_root is not None:
            interp_root = self.represented_interpretation_root
            interp_uuid = bu.uuid_from_string(interp_root.attrib['uuid'])
            interp_part = self.model.part_for_uuid(interp_uuid)
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.find_nested_tags_text(interp_root, ['Citation', 'Title']),
                                       interp_uuid,
                                       content_type = self.model.type_of_part(interp_part),
                                       root = ps_node)

        for patch_index in range(self.patch_count):
            p_node = rqet.SubElement(ps_node, ns['resqml2'] + 'NodePatch')
            p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'NodePatch')
            p_node.text = '\n'

            pi_node = rqet.SubElement(p_node, ns['resqml2'] + 'PatchIndex')
            pi_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
            pi_node.text = str(patch_index)

            ct_node = rqet.SubElement(p_node, ns['resqml2'] + 'Count')
            ct_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            ct_node.text = str(self.patch_ref_list[patch_index][2])

            geom = rqet.SubElement(p_node, ns['resqml2'] + 'Geometry')
            geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
            geom.text = '\n'

            self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

            points_node = rqet.SubElement(geom, ns['resqml2'] + 'Points')
            points_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
            points_node.text = '\n'

            coords = rqet.SubElement(points_node, ns['resqml2'] + 'Coordinates')
            coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            coords.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points_{}'.format(patch_index), root = coords)

        if add_as_part:
            self.model.add_part('obj_PointSetRepresentation', self.uuid, ps_node)
            if add_relationships:
                # todo: add multiple crs'es (one per patch)?
                crs_root = self.model.root_for_uuid(self.crs_uuid)
                self.model.create_reciprocal_relationship(ps_node, 'destinationObject', crs_root, 'sourceObject')
                if self.represented_interpretation_root is not None:
                    self.model.create_reciprocal_relationship(ps_node, 'destinationObject',
                                                              self.represented_interpretation_root, 'sourceObject')
                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(ps_node, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

        return ps_node

    def convert_to_charisma(self, file_name):
        """Output to Charisma 3D interpretation format from a pointset.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name to save to
        """
        #      hznname = self.title.replace(" ","_")
        lines = []
        self.load_all_patches()
        for patch in self.patch_array_list:
            for points in patch:
                lines.append(f"INLINE :\t1 XLINE :\t1\t{points[0]}\t{points[1]}\t{points[2]}\n")
        with open(file_name, 'w') as f:
            for item in lines:
                f.write(item)

    def convert_to_irap(self, file_name):
        """Output to IRAP simple points format from a pointset.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name to save to
        """
        #      hznname = self.title.replace(" ","_")
        lines = []
        self.load_all_patches()
        for patch in self.patch_array_list:
            for points in patch:
                lines.append(f"{points[0]} {points[1]} {points[2]}\n")
        with open(file_name, 'w') as f:
            for item in lines:
                f.write(item)


def concat_polyset_points(closed, coordinates, poly_coords):
    """Concatenates two numpy arrays of coordinates for polylines, omitting the final (duplicated) point in a polyline if the polyline is closed.

    arguments:
        closed (bool): True if polyline is closed
        coordinates (np.array): Coordinates to be concatenated
        poly_coords (np.array): Array to concatenate the given coordinates to

    returns:
        poly_coords (np.array) with additional coordinates concatenated
    """
    if closed and vec.isclose(coordinates[0], coordinates[-1]):
        poly_coords = np.concatenate((poly_coords, coordinates[:-1]))
    else:
        poly_coords = np.concatenate((poly_coords, coordinates))
    return poly_coords



"""simple_lines.py: functions for handling simple lines in relation to a resqml grid."""

version = '19th January 2021'

# line is represented as 2D numpy array of shape (NP, 3): [point index, xyz]

import logging

log = logging.getLogger(__name__)

import numpy as np


def read_lines(filename):
    """Returns a list of line arrays, read from ascii file.

    argument:
       filename (string): the path of the ascii file holding a set of poly-lines

    returns:
       list of numpy arrays, each array representing one poly-line

    notes:
       each line in the file must contain 3 floating point numbers: x, y, z;
       each poly-line must be terminated with a null marker line: 999.0 999.0 999.0
       there is no handling of units; elsewhere they will implicitly be assumed to be
       those of a crs for a grid object
    """

    lines_list = []
    end_of_line = np.array([999.0, 999.0, 999.0])
    try:
        with open(filename, 'r') as fp:
            point_count = 0
            point_list = None
            while True:
                line = fp.readline()
                if len(line) == 0:
                    assert point_count == 0, 'unterminated list of line points at end of file'
                    break
                words = line.split()
                if len(words) == 0:
                    continue  # blank line
                assert len(words) == 3, 'badly formed line (expecting 3 reals)'
                point = np.empty(3)
                for p in range(3):
                    point[p] = float(words[p])
                if np.all(np.isclose(point, end_of_line)):
                    if point_count:
                        lines_list.append(point_list)
                        point_count = 0
                elif point_count:
                    point_list = np.concatenate((point_list, point.reshape((1, 3))), axis = 0)
                    point_count += 1
                else:
                    point_list = np.empty((1, 3))
                    point_list[0, :] = point
                    point_count = 1
    except Exception:
        log.exception('failed to read simple lines from ascii file: ' + filename)
        lines_list = []
    log.info(str(len(lines_list)) + ' line(s) read from file: ' + filename)
    return lines_list


def polygon_line(line, tolerance = 0.001):
    """Returns a copy of the line with the last vertex stripped off if it is close to the first vertex.

    arguments:
       line (numpy array of floats): representation of a poly-line which might be closed (last vertex
          matches first vertex)
       tolerance (float, default = 0.001): the maximum Manhatten distance between two points for them
          to be treated as coincident

    returns:
       numpy array of floats which is either a copy of line, or a copy of line with the last point
       removed
    """

    last = len(line) - 1
    if last < 1:
        return line
    dims = len(line[0])
    manhattan = 0.0
    for d in range(dims):
        manhattan += abs(line[0][d] - line[last][d])
    if manhattan > tolerance:
        return line
    return line[:-1]


def duplicate_vertices_removed(line, tolerance = 0.001):
    """Returns a copy of the line with neighbouring duplicate vertices removed.

    arguments:
       line (2D numpy array of floats): representation of a poly-line
       tolerance (float, default = 0.001): the maximum Manhatten distance between two points for them
          to be treated as coincident

    returns:
       numpy array of floats which is either a copy of line, or a copy of line with some points
       removed

    notes:
       does not treat the line as a closed polyline, use polygon_line() function as well to remove
       duplicated first/last point;
       always preserves first and last point, even if they are identical and there are no other vertices
    """

    assert line.ndim == 2
    if len(line) < 3:
        return line
    dims = line.shape[1]
    whittled = np.zeros(line.shape)
    whittled[0] = line[0]
    c_i = 0
    for i in range(1, len(line)):
        manhattan = 0.0
        for d in range(dims):
            manhattan += abs(line[i, d] - whittled[c_i, d])
        if manhattan > tolerance:
            c_i += 1
            whittled[c_i] = line[i]
    if c_i == len(line) - 1:
        return line
    whittled[c_i] = line[-1]
    return whittled[:c_i + 1]


def nearest_pillars(line_list, grid, ref_k = 0, ref_kp = 0):
    """Finds pillars nearest to each point on each line; returns list of lists of (j0, i0).

    arguments:
       line_list (list of numpy arrays of floats): set of poly-lines, the points of which are used
          to find the nearest pillars in grid
       grid (grid.Grid object): the grid whose pillars are compared with the poly-line vertices
       ref_k (integer, default 0): the reference layer in the grid to compare against the vertices;
          zero based
       ref_kp (integer, default 0): 0 to indicate the top corners of the reference layer, 1 for the base

    returns:
       a list of lists of pairs of integers, each being the (j, i) pillar indices of the nearest pillar
       to the corresponding vertex of the poly-line; zero based indexing

    notes:
       this is a 2D search in the x, y plane; z values are ignored;
       poly-line x, y values must implicitly be in the same crs as the grid's points data
    """

    pillar_list_list = []
    for line in line_list:
        pillar_list = []
        p_count = line.shape[0]
        for p in range(p_count):
            xy = line[p, 0:2]
            ji = grid.nearest_pillar(xy, ref_k0 = ref_k + ref_kp)
            pillar_list.append(ji)
        pillar_list_list.append(pillar_list)
    return pillar_list_list


def nearest_rods(line_list, projection, grid, axis, ref_slice0 = 0, plus_face = False):
    """Finds rods nearest to each point on each line; returns list of lists of (k0, j0) or (k0, i0).

    arguments:
       line_list (list of numpy arrays of floats): set of poly-lines, the points of which are used
          to find the nearest rods in grid
       projection (string): 'xz' or 'yz'
       grid (grid.Grid object): the grid whose cross section points are compared with the poly-line vertices
       axis (string): 'I' or 'J' being the axis removed during slicing
       ref_slice0 (integer, default 0): the reference slice in the grid to compare against the vertices;
          zero based
       plus_face (boolean, default False): which face of the reference slice to use

    returns:
       a list of numpy arrays of pairs of integers, each being the (k, j) or (k, i) indices of the
       nearest rod to the corresponding vertex of the poly-line under projection; zero based indexing

    notes:
       this is a 2D search in the x, z or y, z plane;
       currently limited to unsplit grids without k gaps;
       poly-line x, y, z values must implicitly be in the same crs as the grid's points data
    """

    assert projection in ['xz', 'yz']
    assert axis.upper() in ['J', 'I']

    rod_list_list = []
    for line in line_list:
        line_a = np.array(line)
        rod_list_list.append(grid.nearest_rod(line_a, projection, axis, ref_slice0 = ref_slice0, plus_face = plus_face))
    return rod_list_list


def drape_lines(line_list, pillar_list_list, grid, ref_k = 0, ref_kp = 0, offset = -1.0, snap = False):
    """Roughly drapes lines over grid horizon; draped lines are suitable for 3D visualisation.

    arguments:
       line_list (list of numpy arrays of floats): the undraped poly-lines
       pillar_list_list: (list of lists of pairs of integers): as returned by nearest_pillars()
       grid: (grid.Grid object): the grid to which the poly-lines are to be draped
       ref_k (integer, default 0): the reference layer in the grid to which the lines will be
          draped; zero based
       ref_kp (integer, default 0): 0 to indicate the top corners of the reference layer, 1 for the base
       offset (float, default -1.0): the vertical offset to add to the z value of pillar points; positive
          drapes lines deeper, negative shallower (assumes grid's crs has z increasing downwards)
       snap (boolean, default False): if True, the x & y values of each vertex in the lines are moved to
          match the pillar point; if False, only the z values are adjusted

    returns:
       list of numpy arrays of floats, being the draped equivalents of the line_list

    notes:
       the units of the line list must implicitly be those of the crs for the grid;
       for grids with split coordinate lines (faults), only the primary pillars are currently used;
       the results of this function are intended for 3D visualisation of an indicative nature;
       resulting draped lines may penetrate the grid layer faces depending on the 3D geometry and
       the spacing of the vertices compared to cell sizes
    """

    assert len(line_list) == len(pillar_list_list)

    # note: currently uses primary pillars only: expect draping disasters by faults
    # pe_j = grid.extent_kji[1] + 1
    # pe_i = grid.extent_kji[2] + 1
    pillar_points = grid.horizon_points(ref_k0 = ref_k + ref_kp)

    draped_list = []
    for line, pillar_list in zip(line_list, pillar_list_list):
        drape = line.copy()
        p_count = line.shape[0]
        for p in range(p_count):
            ji = pillar_list[p]
            if snap:
                drape[p, :] = pillar_points[tuple(ji)]
            else:
                drape[p, 2] = pillar_points[tuple(ji)][2]
        drape[:, 2] += offset
        draped_list.append(drape)

    return draped_list


def drape_lines_to_rods(line_list,
                        rod_list_list,
                        projection,
                        grid,
                        axis,
                        ref_slice0 = 0,
                        plus_face = False,
                        offset = -1.0,
                        snap = False):
    """Roughly drapes lines near grid cross section; draped lines are suitable for 3D visualisation.

    arguments:
       line_list (list of numpy arrays of floats): the undraped poly-lines
       rod_list_list: (list of arrays of pairs of integers): as returned by nearest_rods()
       grid: (grid.Grid object): the grid to which the poly-lines are to be draped
       axis (string): 'I' or 'J' being the axis removed during slicing
       ref_slice0 (integer, default 0): the reference slice in the grid to drape to; zero based
       plus_face (boolean, default False): which face of the reference slice to use
       offset (float, default -1.0): the horzontal offset to add to the y or x value of rod points
       snap (boolean, default False): if True, the x & z or y & z values of each vertex in the lines
          are moved to match the rod point; if False, only the y or x values are adjusted

    returns:
       list of numpy arrays of floats, being the draped equivalents of the line_list

    notes:
       the units of the line list must implicitly be those of the crs for the grid;
       currently limited to unsplit grids with no k gaps;
       the results of this function are intended for 3D visualisation of an indicative nature;
       resulting draped lines may penetrate the grid layer faces depending on the 3D geometry and
       the spacing of the vertices compared to cell sizes
    """

    assert projection in ['xz', 'yz']
    assert axis.upper() in ['I', 'J']
    assert len(line_list) == len(rod_list_list)

    x_sect = grid.unsplit_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face)

    draped_list = []
    for line, rod_list in zip(line_list, rod_list_list):
        drape = line.copy()
        p_count = line.shape[0]
        for p in range(p_count):
            kj_or_ki = rod_list[p]
            if snap:
                drape[p, :] = x_sect[tuple(kj_or_ki)]
            elif projection == 'xz':
                drape[p, 1] = x_sect[tuple(kj_or_ki)][1]
            else:
                drape[p, 0] = x_sect[tuple(kj_or_ki)][0]
        if projection == 'xz':
            drape[:, 1] += offset
        else:
            drape[:, 0] += offset
        draped_list.append(drape)

    return draped_list



"""Classes relating to wells."""

__all__ = [
    'MdDatum', 'BlockedWell', 'Trajectory', 'DeviationSurvey', 'WellboreFrame', 'WellboreMarkerFrame', 'WellboreMarker',
    'add_wells_from_ascii_file', 'well_name', 'add_las_to_trajectory', 'add_blocked_wells_from_wellspec',
    'add_logs_from_cellio', 'lookup_from_cellio'
]

from ._md_datum import MdDatum
from ._wellbore_marker import WellboreMarker
from ._wellbore_marker_frame import WellboreMarkerFrame
from ._blocked_well import BlockedWell
from ._trajectory import Trajectory
from ._deviation_survey import DeviationSurvey
from ._wellbore_frame import WellboreFrame

from .well_object_funcs import add_wells_from_ascii_file, well_name, add_las_to_trajectory, \
    add_blocked_wells_from_wellspec, add_logs_from_cellio, lookup_from_cellio

# Set "module" attribute of all public objects to this path. Fixes issue #310
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""prism_grid.py: resqpy PrismGrid and VerticalPrismGrid class module."""

version = '24th November 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.olio.intersection as meet
import resqpy.olio.transmission as rqtr
import resqpy.olio.triangulation as tri
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.property as rqp
import resqpy.surface as rqs
import resqpy.weights_and_measures as wam
from resqpy.unstructured._unstructured_grid import UnstructuredGrid


class PrismGrid(UnstructuredGrid):
    """Class for unstructured grids where every cell is a triangular prism.

    note:
       prism cells are not constrained to have a fixed cross-section, though in practice they often will
    """

    def __init__(self,
                 parent_model,
                 uuid = None,
                 find_properties = True,
                 cache_geometry = False,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a new resqpy PrismGrid object (RESQML UnstructuredGrid with cell shape trisngular prism)

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           uuid (uuid.UUID, optional): if present, the new grid object is populated from the RESQML object
           find_properties (boolean, default True): if True and uuid is present, a
              grid property collection is instantiated as an attribute, holding properties for which
              this grid is the supporting representation
           cache_geometry (boolean, default False): if True and uuid is present, all the geometry arrays
              are loaded into attributes of the new grid object
           title (str, optional): citation title for new grid; ignored if uuid is present
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if uuid is present
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if uuid is present

        returns:
           a newly created PrismGrid object
        """

        super().__init__(parent_model = parent_model,
                         uuid = uuid,
                         find_properties = find_properties,
                         geometry_required = True,
                         cache_geometry = cache_geometry,
                         cell_shape = 'prism',
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            assert grr.grid_flavour(self.root) in ['PrismGrid', 'VerticalPrismGrid']
            self.check_prism()

        self.grid_representation = 'PrismGrid'  #: flavour of grid; not much used

    def check_prism(self):
        """Checks that each cell has 5 faces and each face has 3 or 4 nodes.

        note:
           currently only performs a cursory check, without checking nodes are shared or that there are exactly two
           triangular faces without shared nodes
        """

        assert self.cell_shape == 'prism'
        self.cache_all_geometry_arrays()
        assert self.faces_per_cell_cl is not None and self.nodes_per_face_cl is not None
        assert self.faces_per_cell_cl[0] == 5 and np.all(self.faces_per_cell_cl[1:] - self.faces_per_cell_cl[:-1] == 5)
        nodes_per_face_count = np.empty(self.face_count)
        nodes_per_face_count[0] = self.nodes_per_face_cl[0]
        nodes_per_face_count[1:] = self.nodes_per_face_cl[1:] - self.nodes_per_face_cl[:-1]
        assert np.all(np.logical_or(nodes_per_face_count == 3, nodes_per_face_count == 4))

    # todo: add prism specific methods for centre_point(), volume()


class VerticalPrismGrid(PrismGrid):
    """Class for unstructured grids where every cell is a vertical triangular prism.

    notes:
       vertical prism cells are constrained to have a fixed triangular horzontal cross-section, though top and base
       triangular faces need not be horizontal; edges not involved in the triangular faces must be vertical;
       this is not a native RESQML sub-class but is a resqpy concoction to allow optimisation of some methods;
       face ordering within a cell is also constrained to be top, base, then the three vertical planar quadrilateral
       faces; node ordering within triangular faces is constrained to ensure correspondence of nodes in triangles
       within a column
    """

    def __init__(self,
                 parent_model,
                 uuid = None,
                 find_properties = True,
                 cache_geometry = False,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a new resqpy VerticalPrismGrid object.

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           uuid (uuid.UUID, optional): if present, the new grid object is populated from the RESQML object
           find_properties (boolean, default True): if True and uuid is present, a
              grid property collection is instantiated as an attribute, holding properties for which
              this grid is the supporting representation
           cache_geometry (boolean, default False): if True and uuid is present, all the geometry arrays
              are loaded into attributes of the new grid object
           title (str, optional): citation title for new grid; ignored if uuid is present
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if uuid is present
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if uuid is present

        returns:
           a newly created VerticalPrismGrid object
        """

        self.nk = None  #: number of layers when constructed as a layered grid

        super().__init__(parent_model = parent_model,
                         uuid = uuid,
                         find_properties = find_properties,
                         cache_geometry = cache_geometry,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            assert grr.grid_flavour(self.root) in ['VerticalPrismGrid', 'PrismGrid']
            self.check_prism()
            if 'layer count' in self.extra_metadata:
                self.nk = int(self.extra_metadata['layer count'])

        self.grid_representation = 'VerticalPrismGrid'  #: flavour of grid; not much used

    @classmethod
    def from_surfaces(cls,
                      parent_model,
                      surfaces,
                      column_points = None,
                      column_triangles = None,
                      title = None,
                      originator = None,
                      extra_metadata = {},
                      set_handedness = False):
        """Create a layered vertical prism grid from an ordered list of untorn surfaces.

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           surfaces (list of surface.Surface): list of two or more untorn surfaces ordered from
              shallowest to deepest; see notes
           column_points (2D numpy float array, optional): if present, the xy points to use for
              the grid's triangulation; see notes
           column_triangles (numpy int array of shape (M, 3), optional): if present, indices into the
              first dimension of column_points giving the xy triangulation to use for the grid; see notes
           title (str, optional): citation title for the new grid
           originator (str, optional): name of person creating the grid; defaults to login id
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid

        returns:
           a newly created VerticalPrismGrid object

        notes:
           this method will not work for torn (faulted) surfaces, nor for surfaces with recumbent folds;
           the surfaces may not cross each other, ie. the depth ordering must be consistent over the area;
           the triangular pattern of the columns (in the xy plane) can be specified with the column_points
           and column_triangles arguments;
           if those arguments are None, the first, shallowest, surface is used as a master and determines
           the triangular pattern of the columns;
           where a gravity vector from a node above does not intersect a surface, the point is inherited
           as a copy of the node above and will be NaNs if no surface above has an intersection;
           the Surface class has methods for creating a Surface from a PointSet or a Mesh (RESQML
           Grid2dRepresentation), or for a horizontal plane;
           this class is represented in RESQML as an UnstructuredGridRepresentation – when a resqpy
           class is written for ColumnLayerGridRepresentation, a method will be added to that class to
           convert from a resqpy VerticalPrismGrid
        """

        def find_pair(a, pair):
            # for sorted array a of shape (N, 2) returns index in first axis of a pair

            def frp(a, pair, b, c):
                m = b + ((c - b) // 2)
                assert m < len(a), 'pair not found in sorted array'
                if np.all(a[m] == pair):
                    return m
                assert c > b, 'pair not found in sorted array'
                if a[m, 0] < pair[0]:
                    return frp(a, pair, m + 1, c)
                elif a[m, 0] > pair[0]:
                    return frp(a, pair, b, m)
                elif a[m, 1] < pair[1]:
                    return frp(a, pair, m + 1, c)
                else:
                    return frp(a, pair, b, m)

            return frp(a, pair, 0, len(a))

        assert (column_points is None) == (column_triangles is None)
        assert len(surfaces) > 1
        for s in surfaces:
            assert isinstance(s, rqs.Surface)

        vpg = cls(parent_model, title = title, originator = originator, extra_metadata = extra_metadata)
        assert vpg is not None

        top = surfaces[0]

        # set and check consistency of crs
        vpg.crs_uuid = top.crs_uuid
        for s in surfaces[1:]:
            if not bu.matching_uuids(vpg.crs_uuid, s.crs_uuid):
                # check for equivalence
                assert rqc.Crs(parent_model,
                               uuid = vpg.crs_uuid) == rqc.Crs(parent_model,
                                                               uuid = s.crs_uuid), 'mismatching surface crs'

        # fetch the data for the top surface, to be used as the master for the triangular pattern
        if column_triangles is None:
            top_triangles, top_points = top.triangles_and_points()
            column_edges = top.distinct_edges()  # ordered pairs of node indices
        else:
            top_triangles = column_triangles
            if column_points.shape[1] == 3:
                top_points = column_points
            else:
                top_points = np.zeros((len(column_points), 3))
                top_points[:, :column_points.shape[1]] = column_points
            column_surf = rqs.Surface(parent_model, crs_uuid = vpg.crs_uuid)
            column_surf.set_from_triangles_and_points(column_triangles, column_points)
            column_edges = column_surf.distinct_edges()
        assert top_triangles.ndim == 2 and top_triangles.shape[1] == 3
        assert top_points.ndim == 2 and top_points.shape[1] in [2, 3]
        assert len(top_triangles) > 0
        p_count = len(top_points)
        bad_points = np.zeros(p_count, dtype = bool)

        # setup size of arrays for the vertical prism grid
        column_count = top_triangles.shape[0]
        surface_count = len(surfaces)
        layer_count = surface_count - 1
        column_edge_count = len(column_edges)
        vpg.cell_count = column_count * layer_count
        vpg.node_count = p_count * surface_count
        vpg.face_count = column_count * surface_count + column_edge_count * layer_count
        vpg.nk = layer_count
        if vpg.extra_metadata is None:
            vpg.extra_metadata = {}
        vpg.extra_metadata['layer count'] = vpg.nk

        # setup points with copies of points for top surface, z values to be updated later
        points = np.zeros((surface_count, p_count, 3))
        points[:, :, :] = top_points

        # arrange faces with all triangles first, followed by the vertical quadrilaterals
        vpg.nodes_per_face_cl = np.zeros(vpg.face_count, dtype = int)
        vpg.nodes_per_face_cl[:column_count * surface_count] =  \
           np.arange(3, 3 * column_count * surface_count + 1, 3, dtype = int)
        quad_start = vpg.nodes_per_face_cl[column_count * surface_count - 1] + 4
        vpg.nodes_per_face_cl[column_count * surface_count:] =  \
           np.arange(quad_start, quad_start + 4 * column_edge_count * layer_count, 4)
        assert vpg.nodes_per_face_cl[-1] == 3 * column_count * surface_count + 4 * column_edge_count * layer_count
        # populate nodes per face for triangular faces
        vpg.nodes_per_face = np.zeros(vpg.nodes_per_face_cl[-1], dtype = int)
        for surface_index in range(surface_count):
            vpg.nodes_per_face[surface_index * 3 * column_count : (surface_index + 1) * 3 * column_count] =  \
               top_triangles.flatten() + surface_index * p_count
        # populate nodes per face for quadrilateral faces
        quad_nodes = np.empty((layer_count, column_edge_count, 2, 2), dtype = int)
        for layer in range(layer_count):
            quad_nodes[layer, :, 0, :] = column_edges + layer * p_count
            # reverse order of base pairs to maintain cyclic ordering of nodes per face
            quad_nodes[layer, :, 1, 0] = column_edges[:, 1] + (layer + 1) * p_count
            quad_nodes[layer, :, 1, 1] = column_edges[:, 0] + (layer + 1) * p_count
        vpg.nodes_per_face[3 * surface_count * column_count:] = quad_nodes.flatten()
        assert vpg.nodes_per_face[-1] > 0

        # set up faces per cell
        vpg.faces_per_cell = np.zeros(5 * vpg.cell_count, dtype = int)
        vpg.faces_per_cell_cl = np.arange(5, 5 * vpg.cell_count + 1, 5, dtype = int)
        assert len(vpg.faces_per_cell_cl) == vpg.cell_count
        # set cell top triangle indices
        for layer in range(layer_count):
            # top triangular faces of cells
            vpg.faces_per_cell[5 * layer * column_count : (layer + 1) * 5 * column_count : 5] =  \
               layer * column_count + np.arange(column_count)
            # base triangular faces of cells
            vpg.faces_per_cell[layer * 5 * column_count + 1: (layer + 1) * 5 * column_count : 5] =  \
               (layer + 1) * column_count + np.arange(column_count)
        # todo: some clever numpy indexing to irradicate the following for loop
        for col in range(column_count):
            t_nodes = top_triangles[col]
            for t_edge in range(3):
                a, b = t_nodes[t_edge - 1], t_nodes[t_edge]
                if b < a:
                    a, b = b, a
                edge = find_pair(column_edges, (a, b))  # returns index into first axis of column edges
                # set quadrilateral faces of cells in column, for this edge
                vpg.faces_per_cell[5 * col + t_edge + 2 : 5 * vpg.cell_count : 5 * column_count] =  \
                   np.arange(column_count * surface_count + edge, vpg.face_count, column_edge_count)
        # check full population of faces_per_cell (face zero is a top triangle, only used once)
        assert np.count_nonzero(vpg.faces_per_cell) == vpg.faces_per_cell.size - 1

        vpg.cell_face_is_right_handed = np.ones(len(vpg.faces_per_cell), dtype = bool)
        if set_handedness:
            # TODO: set handedness correctly and make default for set_handedness True
            raise NotImplementedError('code not written to set handedness for vertical prism grid from surfaces')

        # instersect gravity vectors from column points with other surfaces, and update z values in points
        gravity = np.zeros((p_count, 3))
        gravity[:, 2] = 1.0  # up/down does not matter for the intersection function used below
        start = 1 if top_triangles is None else 0
        for surf in range(start, surface_count):
            surf_triangles, surf_points = surfaces[surf].triangles_and_points()
            intersects = meet.line_set_triangles_intersects(top_points, gravity, surf_points[surf_triangles])
            single_intersects = meet.last_intersects(intersects)  # will be triple NaN where no intersection occurs
            # inherit point from surface above where no intersection has occurred
            nan_lines = np.isnan(single_intersects[:, 0])
            if surf == 0:
                # allow NaN entries to handle unused distant circumcentres in Voronoi graph data
                # assert not np.any(nan_lines), 'top surface does not cover all column points'
                single_intersects[nan_lines] = np.NaN
            else:
                single_intersects[nan_lines] = points[surf - 1][nan_lines]
            # populate z values for layer of points
            points[surf, :, 2] = single_intersects[:, 2]

        vpg.points_cached = points.reshape((-1, 3))
        assert np.all(vpg.nodes_per_face < len(vpg.points_cached))

        return vpg

    @classmethod
    def from_seed_points_and_surfaces(cls,
                                      parent_model,
                                      seed_xy,
                                      surfaces,
                                      area_of_interest,
                                      title = None,
                                      originator = None,
                                      extra_metadata = {},
                                      set_handedness = False):
        """Create a layered vertical prism grid from seed points and an ordered list of untorn surfaces.

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           seed_xy (numpy float array of shape (N, 2 or 3)): the xy locations of seed points
           surfaces (list of surface.Surface): list of two or more untorn surfaces ordered from
              shallowest to deepest; see notes
           area_of_interest (closed convex Polyline): the frontier polygon in the xy plane
           title (str, optional): citation title for the new grid
           originator (str, optional): name of person creating the grid; defaults to login id
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid

        returns:
           a newly created VerticalPrismGrid object

        notes:
           the triangular pattern of the columns (in the xy plane) is constructed as a re-triangulation
           of the Voronoi diagram of the Delauney triangulation of the seed points;
           the seed points may be xy or xyz data but z is ignored;
           this method will not work for torn (faulted) surfaces, nor for surfaces with recumbent folds;
           the surfaces may not cross each other, ie. the depth ordering must be consistent over the area;
           all the seed points must lie wholly within the area of interest;
           where a gravity vector from a node above does not intersect a surface, the point is inherited
           as a copy of the node above (the topmost surface must cover the entire area of interest);
           the Surface class has methods for creating a Surface from a PointSet or a Mesh (RESQML
           Grid2dRepresentation), or for a horizontal plane
        """

        assert seed_xy.ndim == 2 and seed_xy.shape[1] in [2, 3]
        assert area_of_interest.isclosed and area_of_interest.is_convex()

        # compute Delauney triangulation
        delauney_t, hull_indices = tri.dt(seed_xy, return_hull = True, algorithm = 'scipy')

        # construct Voronoi graph
        voronoi_points, voronoi_indices = tri.voronoi(seed_xy, delauney_t, hull_indices, area_of_interest)

        # re-triangulate Voronoi cells
        points, triangles = tri.triangulated_polygons(voronoi_points, voronoi_indices, centres = seed_xy[:, :2])

        vpg = cls.from_surfaces(parent_model,
                                surfaces,
                                column_points = points,
                                column_triangles = triangles,
                                title = title,
                                originator = originator,
                                extra_metadata = extra_metadata,
                                set_handedness = set_handedness)

        # TODO: store information relating columns to seed points, and seed points to top layer points?

        return vpg

    def column_count(self):
        """Returns the number of columns in the grid."""

        assert self.nk > 1, 'no layer information set for vertical prism grid'
        n_col, remainder = divmod(self.cell_count, self.nk)
        assert remainder == 0, 'code failure for vertical prism grid: cell and layer counts not compatible'
        return n_col

    def cell_centre_point(self, cell = None):
        """Returns centre point of a single cell (or all cells) calculated as the mean position of its 6 nodes.

        arguments:
           cell (int): the index of the cell for which the centre point is required

        returns:
           numpy float array of shape (3,) being the xyz location of the centre point of the cell
        """

        cp = self.corner_points(cell = cell)
        return np.mean(cp.reshape((-1, 6, 3)), axis = (1))

    def corner_points(self, cell = None):
        """Returns corner points for all cells or for a single cell.

        arguments:
           cell (int, optional): cell index of single cell for which corner points are required;
              if None, corner points are returned for all cells

        returns:
           numpy float array of shape (2, 3, 3) being xyz points for top & base points for one cell, or
           numpy float array of shape (N, 2, 3, 3) being xyz points for top & base points for all cells
        """

        if hasattr(self, 'array_corner_points'):
            if cell is None:
                return self.array_corner_points
            return self.array_corner_points[cell]

        p = self.points_ref()
        if cell is None:
            cp = np.empty((self.cell_count, 2, 3, 3), dtype = float)
            top_fi = self.faces_per_cell.reshape((-1, 5))[:, 0]
            top_npfi_start = np.where(top_fi == 0, 0, self.nodes_per_face_cl[top_fi - 1])
            top_npfi_end = self.nodes_per_face_cl[top_fi]
            base_fi = self.faces_per_cell.reshape((-1, 5))[:, 1]
            base_npfi_start = self.nodes_per_face_cl[base_fi - 1]
            base_npfi_end = self.nodes_per_face_cl[base_fi]
            for cell in range(self.cell_count):
                cp[cell, 0] = p[self.nodes_per_face[top_npfi_start[cell]:top_npfi_end[cell]]]
                cp[cell, 1] = p[self.nodes_per_face[base_npfi_start[cell]:base_npfi_end[cell]]]
            self.array_corner_points = cp
            return cp

        top_fi, base_fi = self.faces_per_cell.reshape((-1, 5))[cell, :2]
        top_npfi_start = 0 if top_fi == 0 else self.nodes_per_face_cl[top_fi - 1]
        base_npfi_start = self.nodes_per_face_cl[base_fi - 1]
        cp = np.empty((2, 3, 3), dtype = float)
        cp[0] = p[self.nodes_per_face[top_npfi_start:self.nodes_per_face_cl[top_fi]]]
        cp[1] = p[self.nodes_per_face[base_npfi_start:self.nodes_per_face_cl[base_fi]]]
        return cp

    def thickness(self, cell = None):
        """Returns array of thicknesses of all cells or a single cell.

        note:
           units are z units of crs used by this grid
        """

        if hasattr(self, 'array_thickness'):
            if cell is None:
                return self.array_thickness
            return self.array_thickness[cell]

        cp = self.corner_points(cell = cell)

        if cell is None:
            thick = np.mean(cp[:, 1, :, 2] - cp[:, 0, :, 2], axis = -1)
            self.array_thickness = thick
        else:
            thick = np.mean(cp[1, :, 2] - cp[0, :, 2])

        return thick

    def top_faces(self):
        """Returns the global face indices for the top triangular faces of the top layer of cells."""

        return self.faces_per_cell[:self.column_count() * 5].reshape(-1, 5)[:, 0]

    def triangulation(self):
        """Returns triangulation used by this vertical prism grid.

        returns:
           numpy int array of shape (M, 3) being the indices into the grid points of the triangles forming
           the top faces of the top layer of cells of the grid

        notes:
           use points_ref() to access the full points data;
           the order of the first axis of the returned values will match the order of the columns
           within cell related data
        """

        n_col = self.column_count()
        top_face_indices = self.top_faces()
        top_nodes = np.empty((n_col, 3), dtype = int)
        for column in range(n_col):  # also used as cell number
            top_nodes[column] = self.node_indices_for_face(top_face_indices[column])
        return top_nodes

    def triple_horizontal_permeability(self, primary_k, orthogonal_k, primary_azimuth = 0.0):
        """Returns array of triple horizontal permeabilities derived from a pair of permeability properties.

        arguments:
           primary_k (numpy float array of shape (N,) being the primary horizontal permeability for each cell
           orthogonal_k (numpy float array of shape (N,) being the horizontal permeability for each cell in a
              direction orthogonal to the primary permeability
           primary_azimuth (float or numpy float array of shape (N,), default 0.0): the azimuth(s) of the
              primary permeability, in degrees compass bearing

        returns:
           numpy float array of shape (N, 3) being the triple horizontal permeabilities applicable to half
           cell horizontal transmissibilities

        notes:
           this method should be used to generate horizontal permeabilities for use in transmissibility
           calculations if starting from an anisotropic permeability defined by a pair of locally
           orthogonal values; resulting values will be locally bounded by the pair of source values;
           the order of the 3 values per cell follows the node indices for the top triangle, for the
           opposing vertical face; no net to gross ratio modification is applied here;
        """

        assert primary_k.size == self.cell_count and orthogonal_k.size == self.cell_count
        azimuth_is_constant = isinstance(primary_azimuth, float) or isinstance(primary_azimuth, int)
        if not azimuth_is_constant:
            assert primary_azimuth.size == self.cell_count

        p = self.points_ref()
        t = self.triangulation()
        # mid point of triangle edges
        m = np.empty((t.shape[0], 3, 3), dtype = float)
        for e in range(3):
            m[:, e, :] = 0.5 * (p[t[:, e]] + p[t[:, e - 1]])
        # cell centre points
        c = self.centre_point()
        # todo: decide which directions to use: cell centre to face centre; or face normal?
        # vectors with direction cell centre to face centre (for one layer only)
        v = m - c.reshape((self.nk, -1, 1, 3))[0]
        # compass directions of those vectors
        a = vec.azimuths(v)
        # vector directions relative to primary permeability direction (per cell)
        if azimuth_is_constant:
            # work with one layer only
            ap_rad = np.radians(a - primary_azimuth)
        else:
            # value per face per cell in whole grid
            ap_rad = np.radians(
                np.repeat(a.reshape((1, -1, 3)), self.nk, axis = 0).reshape((-1, 3)) - primary_azimuth.reshape((-1, 1)))
        cos_ap = np.cos(ap_rad)
        sin_ap = np.sin(ap_rad)
        cos_sqr_ap = cos_ap * cos_ap
        sin_sqr_ap = sin_ap * sin_ap
        if azimuth_is_constant:
            cos_sqr_ap = np.repeat(cos_sqr_ap.reshape(1, -1, 3), self.nk, axis = 0).reshape((-1, 3))
            sin_sqr_ap = np.repeat(sin_sqr_ap.reshape(1, -1, 3), self.nk, axis = 0).reshape((-1, 3))
        # local elliptical permeability projected in directions of azimuths
        k = np.sqrt((primary_k * primary_k).reshape((-1, 1)) * cos_sqr_ap +
                    (orthogonal_k * orthogonal_k).reshape((-1, 1)) * sin_sqr_ap)

        return k

    def half_cell_transmissibility(self, use_property = True, realization = None, tolerance = 1.0e-6):
        """Returns (and caches if realization is None) half cell transmissibilities for this vertical prism grid.

        arguments:
           use_property (boolean, default True): if True, the grid's property collection is inspected for
              a possible half cell transmissibility array and if found, it is used instead of calculation
           realization (int, optional) if present, only a property with this realization number will be used
           tolerance (float, default 1.0e-6): minimum half axis length below which the transmissibility
              will be deemed uncomputable (for the axis in question); NaN values will be returned (not Inf);
              units are implicitly those of the grid's crs length units

        returns:
           numpy float array of shape (N, 5) where N is the number of cells in the grid and the 5 covers
           the faces of a cell in the order of the faces_per_cell data;
           units will depend on the length units of the coordinate reference system for the grid;
           the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m and ft respectively

        notes:
           this method does not write to hdf5, nor create a new property or xml;
           if realization is None, a grid attribute cached array will be used;
           tolerance will only be used if the half cell transmissibilities are actually computed
        """

        if realization is None and hasattr(self, 'array_half_cell_t'):
            return self.array_half_cell_t

        half_t = None
        pc = self.property_collection

        if use_property:
            half_t = pc.single_array_ref(property_kind = 'transmissibility',
                                         realization = realization,
                                         continuous = True,
                                         count = 1,
                                         indexable = 'faces per cell')
            if half_t:
                assert half_t.size == 5 * self.cell_count
                half_t = half_t.reshape(self.cell_count, 5)

        if half_t is None:
            # note: properties must be identifiable in property_collection
            # TODO: gather property arrays, deriving tri-permeablities
            # make sub-collection of permeability properties
            ppc = rqp.selective_version_of_collection(pc,
                                                      continuous = True,
                                                      realization = realization,
                                                      property_kind = 'permeability rock')
            assert ppc.number_of_parts() > 0, 'no permeability properties available for vertical prism grid'

            # look for a triple permeability; if present, assume to be per face horizontal permeabilities
            triple_perm_horizontal = ppc.single_array_ref(count = 3, indexable = 'cells')

            if triple_perm_horizontal is None:
                # look for horizontal isotropic permeability
                iso_h_part = None
                if ppc.number_of_parts() == 1:
                    iso_h_part = ppc.parts()[0]
                    assert ppc.indexable_for_part(
                        iso_h_part) == 'cells', 'single permeability property is not for cells'
                else:
                    iso_h_part = ppc.singleton(facet_type = 'direction',
                                               facet = 'horizontal',
                                               count = 1,
                                               indexable = 'cells')
                if iso_h_part is not None:
                    triple_perm_horizontal = np.repeat(ppc.cached_part_array_ref(iso_h_part), 3).reshape((-1, 3))
                else:
                    # look for horizontal permeability field properties and derive triple perm array
                    primary_k = ppc.single_array_ref(facet_type = 'direction',
                                                     facet = 'primary',
                                                     count = 1,
                                                     indexable = 'cells')
                    orthogonal_k = ppc.single_array_ref(facet_type = 'direction',
                                                        facet = 'orthogonal',
                                                        count = 1,
                                                        indexable = 'cells')
                    assert primary_k is not None and orthogonal_k is not None,  \
                       'failed to identify horizontal permeability properties for vertical prism grid'
                    # todo: add extra metadata or further faceting to be sure of finding correct property
                    azimuth_part = pc.singleton(property_kind = 'plane angle',
                                                facet_type = 'direction',
                                                facet = 'primary',
                                                realization = realization,
                                                continuous = True,
                                                count = 1,
                                                indexable = 'cells')
                    if azimuth_part is None:
                        primary_azimuth = 0.0  # default to north
                    else:
                        primary_azimuth = pc.cached_part_array_ref(azimuth_part)
                        azi_uom = pc.uom_for_part(azimuth_part)
                        if azi_uom is not None and azi_uom != 'dega':
                            primary_azimuth = wam.convert(primary_azimuth, azi_uom, 'dega', quantity = 'plane angle')
                    triple_perm_horizontal = self.triple_horizontal_permeability(primary_k, orthogonal_k,
                                                                                 primary_azimuth)

            perm_k = ppc.single_array_ref(facet_type = 'direction', facet = 'K', count = 1, indexable = 'cells')

            ntg = pc.single_array_ref(property_kind = 'net to gross ratio',
                                      realization = realization,
                                      continuous = True,
                                      count = 1,
                                      indexable = 'cells')

            half_t = rqtr.half_cell_t_vertical_prism(self,
                                                     triple_perm_horizontal = triple_perm_horizontal,
                                                     perm_k = perm_k,
                                                     ntg = ntg)

        if realization is None:
            self.array_half_cell_t = half_t

        return half_t
